"""
The MIT License (MIT)

Copyright (c) 2015 Rossen Georgiev <rossen@rgp.io>
Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

This contains a copy of
https://github.com/ValvePython/steam/blob/master/steam/steamid.py
"""

from __future__ import annotations

import abc
import asyncio
import re
from collections.abc import Coroutine
from datetime import datetime
from typing import TYPE_CHECKING, Any, TypeVar

import attr
from typing_extensions import Final, Protocol, Self, TypedDict, runtime_checkable

from .badge import FavouriteBadge, UserBadges
from .comment import Comment
from .enums import *
from .errors import WSException
from .game import Game, StatefulGame, UserGame, WishlistGame
from .iterators import AsyncIterator, CommentsIterator
from .models import URL, Ban
from .profile import *
from .trade import Inventory
from .utils import (
    _INVITE_HEX,
    _INVITE_MAPPING,
    InstanceType,
    Intable,
    TypeType,
    UniverseType,
    cached_slot_property,
    id64_from_url,
    make_id64,
)

if TYPE_CHECKING:
    from aiohttp import ClientSession

    from .clan import Clan
    from .group import Group
    from .http import StrOrURL
    from .image import Image
    from .message import Authors
    from .protobufs.chat import Mentions
    from .state import ConnectionState
    from .user import User

__all__ = (
    "SteamID",
    "Message",
    "Channel",
)

C = TypeVar("C", bound="Commentable")
M_co = TypeVar("M_co", bound="Message", covariant=True)


class UserDict(TypedDict):
    steamid: str
    personaname: str
    primaryclanid: str
    profileurl: str
    realname: str
    # enums
    communityvisibilitystate: int
    profilestate: int
    commentpermission: int
    personastate: int
    personastateflags: int
    # avatar
    avatar: str
    avatarmedium: str
    avatarfull: str
    avatarhash: str
    # country stuff
    loccountrycode: str
    locstatecode: int
    loccityid: int
    # game stuff
    gameid: str  # game app id
    gameextrainfo: str  # game name
    # unix timestamps
    timecreated: int
    lastlogoff: int
    # we pass these ourselves
    last_logon: int
    last_seen_online: int


class SteamID(metaclass=abc.ABCMeta):
    """Convert a Steam ID between its various representations.

    Note
    ----
    See :func:`steam.utils.make_id64` for the full parameter list.
    """

    __slots__ = ("__BASE",)

    def __init__(
        self,
        id: Intable = 0,
        type: TypeType | None = None,
        universe: UniverseType | None = None,
        instance: InstanceType | None = None,
    ):
        self.__BASE: Final = make_id64(id, type, universe, instance)

    def __int__(self) -> int:
        return self.__BASE

    def __eq__(self, other: Any) -> bool:
        try:
            return self.__BASE == int(other)
        except (TypeError, ValueError):
            return NotImplemented

    def __str__(self) -> str:
        return str(self.__BASE)

    def __hash__(self) -> int:
        return hash(self.__BASE)

    def __repr__(self) -> str:
        return f"SteamID(id={self.id}, type={self.type}, universe={self.universe}, instance={self.instance})"

    @property
    def instance(self) -> int:
        """The instance of the SteamID."""
        return (self.__BASE >> 32) & 0xFFFFF

    @property
    def type(self) -> Type:
        """The Steam type of the SteamID."""
        return Type((self.__BASE >> 52) & 0xF)

    @property
    def universe(self) -> Universe:
        """The Steam universe of the SteamID."""
        return Universe((self.__BASE >> 56) & 0xFF)

    @property
    def id64(self) -> int:
        """The SteamID's 64 bit ID."""
        return self.__BASE

    @property
    def id(self) -> int:
        """The SteamID's 32 bit ID."""
        return self.__BASE & 0xFFFFFFFF

    @property
    def id2(self) -> str:
        """The SteamID's ID 2.

        e.g ``STEAM_1:0:1234``.
        """
        return f"STEAM_{self.universe.value}:{self.id % 2}:{self.id >> 1}"

    @property
    def id2_zero(self) -> str:
        """The SteamID's ID 2 accounted for bugged GoldSrc and Orange Box games.

        Note
        ----
        In these games the accounts :attr:`universe`, ``1`` for :class:`.Type.Public`, should be the ``X`` component of
        ``STEAM_X:0:1234`` however, this was bugged and the value of ``X`` was ``0``.

        e.g ``STEAM_0:0:1234``.
        """
        return self.id2.replace("_1", "_0")

    @property
    def id3(self) -> str:
        """The SteamID's ID 3.

        e.g ``[U:1:1234]``.
        """
        type_char = TypeChar(self.type).name
        instance = None

        if self.type in (Type.AnonGameServer, Type.Multiseat):
            instance = self.instance
        elif self.type == Type.Individual:
            if self.instance != 1:
                instance = self.instance
        elif self.type == Type.Chat:
            if self.instance & InstanceFlag.Clan:
                type_char = "c"
            elif self.instance & InstanceFlag.Lobby:
                type_char = "L"
            else:
                type_char = "T"

        parts = [type_char, str(self.universe.value), str(self.id)]

        if instance is not None:
            parts.append(str(instance))

        return f"[{':'.join(parts)}]"

    @property
    def invite_code(self) -> str | None:
        """The SteamID's invite code in the s.team invite code format.

        e.g. ``cv-dgb``.
        """
        if self.type == Type.Individual and self.is_valid():
            invite_code = re.sub(f"[{_INVITE_HEX}]", lambda x: _INVITE_MAPPING[x.group()], f"{self.id:x}")
            split_idx = len(invite_code) // 2
            return invite_code if split_idx == 0 else f"{invite_code[:split_idx]}-{invite_code[split_idx:]}"

    @property
    def invite_url(self) -> str | None:
        """The SteamID's full invite code URL.

        e.g ``https://s.team/p/cv-dgb``.
        """
        code = self.invite_code
        return f"https://s.team/p/{code}" if code else None

    @property
    def community_url(self) -> str | None:
        """The SteamID's community url.

        e.g https://steamcommunity.com/profiles/123456789.
        """
        suffix = {
            Type.Individual: "profiles",
            Type.Clan: "gid",
        }
        try:
            return f"https://steamcommunity.com/{suffix[self.type]}/{self.id64}"
        except KeyError:
            return None

    def is_valid(self) -> bool:
        """Whether or not the SteamID is valid."""
        if self.type == Type.Invalid or self.type >= Type.Max:
            return False

        if self.universe == Universe.Invalid or self.universe >= Universe.Max:
            return False

        if self.type == Type.Individual and (self.id == 0 or self.instance > 4):
            return False

        if self.type == Type.Clan and (self.id == 0 or self.instance != 0):
            return False

        if self.type == Type.GameServer and self.id == 0:
            return False

        if self.type == Type.AnonGameServer and self.id == 0 and self.instance == 0:
            return False

        return True

    @staticmethod
    async def from_url(url: StrOrURL, session: ClientSession | None = None) -> SteamID | None:
        """A helper function creates a SteamID instance from a Steam community url.

        Note
        ----
        See :func:`id64_from_url` for the full parameter list.
        """
        id64 = await id64_from_url(url, session)
        return SteamID(id64) if id64 else None


class Commentable(Protocol):
    """A mixin that implements commenting functionality"""

    __slots__ = ()
    _state: ConnectionState

    @property
    @abc.abstractmethod
    def _commentable_kwargs(self) -> dict[str, Any]:
        raise NotImplementedError

    async def fetch_comment(self, id: int) -> Comment[Self]:
        """Fetch a comment by its ID.

        Parameters
        ----------
        id
            The ID of the comment to fetch.
        """
        comment = await self._state.fetch_comment(self, id)
        return Comment(
            self._state,
            id=comment.id,
            content=comment.content,
            created_at=datetime.utcfromtimestamp(comment.timestamp),
            author=await self._state._maybe_user(comment.author_id64),
            owner=self,
        )

    async def comment(self, content: str, *, subscribe: bool = True) -> Comment[Self]:
        """Post a comment to a comments section.

        Parameters
        ----------
        content
            The message to add to the comment section.
        subscribe
            Whether to subscribe to notifications on any future activity in this comment's thread.

        Returns
        -------
        The created comment.
        """
        return await self._state.post_comment(self, content, subscribe)

    def comments(
        self,
        oldest_first: bool = False,
        limit: int | None = None,
        before: datetime | None = None,
        after: datetime | None = None,
    ) -> CommentsIterator[Self]:
        """An :class:`~steam.iterators.AsyncIterator` for accessing a comment section's :class:`~steam.Comment` objects.

        Examples
        --------

        Usage:

        .. code-block:: python3

            async for comment in commentable.comments(limit=10):
                print("Author:", comment.author, "Said:", comment.content)

        Flattening into a list:

        .. code-block:: python3

            comments = await commentable.comments(limit=50).flatten()
            # comments is now a list of Comment

        All parameters are optional.

        Parameters
        ----------
        oldest_first
            Whether or not to request comments with the oldest comments first or last. Defaults to ``False``.
        limit
            The maximum number of comments to search through.
            Default is ``None`` which will fetch the clan's entire comments section.
        before
            A time to search for comments before.
        after
            A time to search for comments after.

        Yields
        ---------
        :class:`~steam.Comment`
        """
        return CommentsIterator(
            oldest_first=oldest_first, owner=self, state=self._state, limit=limit, before=before, after=after
        )


class BaseUser(SteamID, Commentable):
    """An ABC that details the common operations on a Steam user.
    The following classes implement this ABC:

        - :class:`~steam.User`
        - :class:`~steam.ClientUser`

    .. container:: operations

        .. describe:: x == y

            Checks if two users are equal.

        .. describe:: str(x)

            Returns the user's name.

    Attributes
    ----------
    name
        The user's username.
    state
        The current persona state of the account (e.g. LookingToTrade).
    game
        The Game instance attached to the user. Is ``None`` if the user isn't in a game or one that is recognised by the
        api.
    primary_clan
        The primary clan the User displays on their profile.

        Note
        ----
        This can be lazily awaited to get more attributes of the clan.

    avatar_url
        The avatar url of the user. Uses the large (184x184 px) image url.
    real_name
        The user's real name defined by them. Could be ``None``.
    created_at
        The time at which the user's account was created. Could be ``None``.
    last_logoff
        The last time the user logged into steam. Could be None (e.g. if they are currently online).
    country
        The country code of the account. Could be ``None``.
    flags
        The persona state flags of the account.
    """

    __slots__ = (
        "name",
        "game",
        "state",
        "flags",
        "country",
        "primary_clan",
        "trade_url",
        "real_name",
        "avatar_url",
        "last_seen_online",
        "created_at",
        "last_logoff",
        "last_logon",
        "privacy_state",
        "community_url",
        "_is_commentable",
        "_setup_profile",
        "_level",
        "_state",
    )

    def __init__(self, state: ConnectionState, data: UserDict):
        super().__init__(data["steamid"])
        self._state = state
        self.name: str
        self.real_name: str | None = None
        self.community_url: str | None = None
        self.avatar_url: str | None = None  # TODO make this a property and add avatar hash
        self.primary_clan: Clan | None = None
        self.country: str | None = None
        self.created_at: datetime | None = None
        self.last_logoff: datetime | None = None
        self.last_logon: datetime | None = None
        self.last_seen_online: datetime | None = None
        self.game: StatefulGame | None = None
        self.state: PersonaState | None = None
        self.flags: PersonaStateFlag | None = None
        self.privacy_state: CommunityVisibilityState | None = None
        self._update(data)

    def _update(self, data: UserDict) -> None:
        self.name = data["personaname"]
        self.real_name = data.get("realname") or self.real_name
        self.community_url = data.get("profileurl") or super().community_url
        self.avatar_url = data.get("avatarfull") or self.avatar_url
        self.trade_url = URL.COMMUNITY / f"tradeoffer/new/?partner={self.id}"
        from .clan import Clan  # circular import

        self.primary_clan = (
            Clan(self._state, data["primaryclanid"]) if "primaryclanid" in data else self.primary_clan  # type: ignore
        )
        self.country = data.get("loccountrycode") or self.country
        self.created_at = datetime.utcfromtimestamp(data["timecreated"]) if "timecreated" in data else self.created_at
        self.last_logoff = datetime.utcfromtimestamp(data["lastlogoff"]) if "lastlogoff" in data else self.last_logoff
        self.last_logon = datetime.utcfromtimestamp(data["last_logon"]) if "last_logon" in data else self.last_logon
        self.last_seen_online = (
            datetime.utcfromtimestamp(data["last_seen_online"]) if "last_seen_online" in data else self.last_seen_online
        )
        self.game = (
            StatefulGame(self._state, name=data["gameextrainfo"], id=data["gameid"]) if "gameid" in data else self.game
        )
        self.state = PersonaState.try_value(data.get("personastate", 0)) or self.state
        self.flags = PersonaStateFlag.try_value(data.get("personastateflags", 0)) or self.flags
        self.privacy_state = CommunityVisibilityState.try_value(data.get("communityvisibilitystate", 0))
        self._is_commentable = bool(data.get("commentpermission"))
        self._setup_profile = bool(data.get("profilestate"))

    def __repr__(self) -> str:
        attrs = ("name", "state", "id", "type", "universe", "instance")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"

    def __str__(self) -> str:
        return self.name

    def __del__(self):
        self._state._users.pop(self.id64, None)

    @property
    def _commentable_kwargs(self) -> dict[str, Any]:
        return {
            "id64": self.id64,
            "thread_type": 10,
        }

    @property
    def mention(self) -> str:
        """The string used to mention the user in chat."""
        return f"[mention={self.id}]@{self.name}[/mention]"

    async def inventory(self, game: Game) -> Inventory:
        """Fetch a user's :class:`~steam.Inventory` for trading.

        Parameters
        -----------
        game
            The game to fetch the inventory for.

        Raises
        ------
        :exc:`~steam.Forbidden`
            The user's inventory is private.
        """
        resp = await self._state.http.get_user_inventory(self.id64, game.id, game.context_id)
        return Inventory(state=self._state, data=resp, owner=self, game=game)

    async def friends(self) -> list[User]:
        """Fetch the list of the users friends."""
        friends = await self._state.http.get_friends(self.id64)
        return [self._state._store_user(friend) for friend in friends]

    async def games(self) -> list[UserGame]:
        r"""Fetches the :class:`~steam.Game`\s the user owns."""
        data = await self._state.http.get_user_games(self.id64)
        games = data["response"].get("games", [])
        return [UserGame(self._state, game) for game in games]

    async def clans(self) -> list[Clan]:
        r"""Fetches a list of :class:`~steam.Clan`\s the user is in."""
        clans = []

        async def getter(gid: int) -> None:
            try:
                clan = await self._state.client.fetch_clan(gid)
            except WSException as exc:
                if exc.code == Result.RateLimitExceeded:
                    await asyncio.sleep(20)
                    return await getter(gid)
                raise
            else:
                clans.append(clan)

        resp = await self._state.http.get_user_clans(self.id64)
        for clan in resp["response"]["groups"]:
            await getter(int(clan["gid"]))
        return clans

    async def bans(self) -> Ban:
        r"""Fetches the user's :class:`.Ban`\s."""
        resp = await self._state.http.get_user_bans(self.id64)
        resp = resp["players"][0]
        resp["EconomyBan"] = resp["EconomyBan"] != "none"
        return Ban(data=resp)

    async def badges(self) -> UserBadges:
        r"""Fetches the user's :class:`.UserBadges`\s."""
        resp = await self._state.http.get_user_badges(self.id64)
        return UserBadges(self._state, data=resp["response"])

    async def level(self) -> int:
        """Fetches the user's level if your account is premium, otherwise it's cached."""
        if self._state.http.api_key is not None:
            resp = await self._state.http.get_user_level(self.id64)
            return resp["response"]["player_level"]
        return self._level

    async def wishlist(self) -> list[WishlistGame]:
        r"""Get the :class:`.WishlistGame`\s the user has on their wishlist."""
        data = await self._state.http.get_wishlist(self.id64)
        return [WishlistGame(self._state, id=id, data=game_info) for id, game_info in data.items()]

    async def favourite_badge(self) -> FavouriteBadge | None:
        """The user's favourite badge"""
        badge = await self._state.fetch_user_favourite_badge(self.id64)
        if not badge.has_favorite_badge:
            return

        return FavouriteBadge(
            id=UserBadge.try_value(badge.badgeid),
            item_id=badge.communityitemid,
            type=badge.item_type,
            border_colour=badge.border_color,
            game=StatefulGame(self._state, id=badge.appid) if badge.appid else None,
            level=badge.level,
        )

    async def equipped_profile_items(self) -> EquippedProfileItems:
        """The user's equipped profile items."""
        items = await self._state.fetch_user_equipped_profile_items(self.id64)
        return EquippedProfileItems(
            background=ProfileItem(self._state, items.profile_background) if items.profile_background else None,
            mini_profile_background=ProfileItem(self._state, items.mini_profile_background)
            if items.mini_profile_background
            else None,
            avatar_frame=ProfileItem(self._state, items.avatar_frame) if items.avatar_frame else None,
            animated_avatar=ProfileItem(self._state, items.animated_avatar) if items.animated_avatar else None,
            modifier=ProfileItem(self._state, items.profile_modifier) if items.profile_modifier else None,
        )

    async def profile_info(self) -> ProfileInfo:
        """The user's profile info."""
        info = await self._state.fetch_user_profile_info(self.id64)
        return ProfileInfo(
            created_at=datetime.utcfromtimestamp(info.time_created),
            real_name=info.real_name or None,
            city_name=info.city_name or None,
            state_name=info.state_name or None,
            country_name=info.country_name or None,
            headline=info.headline or None,
            summary=info.summary,
        )

    async def profile(self) -> Profile:
        """Fetch a user's entire profile information.

        Note
        ----
        This calls all the ``profile_x`` functions to return a Profile object which has all the info set.
        """

        coros = [
            self.equipped_profile_items(),
            self.profile_info(),
        ]

        if hasattr(self, "profile_items"):
            coros.append(self.profile_items())

        profiles: tuple[EquippedProfileItems, ProfileInfo] | tuple[
            EquippedProfileItems, ProfileInfo, OwnedProfileItems
        ] = await asyncio.gather(*coros)

        return Profile(*profiles)

    def is_commentable(self) -> bool:
        """Specifies if the user's account is able to be commented on."""
        return self._is_commentable

    def is_private(self) -> bool:
        """Specifies if the user has a private profile."""
        return self.privacy_state == CommunityVisibilityState.Private

    def has_setup_profile(self) -> bool:
        """Specifies if the user has a setup their profile."""
        return self._setup_profile

    async def is_banned(self) -> bool:
        """Specifies if the user is banned from any part of Steam.

        This is equivalent to:

        .. code-block:: python3

            bans = await user.bans()
            bans.is_banned()
        """
        bans = await self.bans()
        return bans.is_banned()

    @classmethod
    def _patch_without_api(cls) -> None:
        import functools

        def __init__(self, state: ConnectionState, data: dict[str, Any]) -> None:
            super().__init__(data["steamid"])
            self._state = state
            self.name = data["persona_name"]
            self.avatar_url = data.get("avatar_url") or self.avatar_url

            self.real_name = NotImplemented
            self.trade_url = URL.COMMUNITY / f"tradeoffer/new?partner={self.id}"
            self.primary_clan = NotImplemented
            self.country = NotImplemented
            self.created_at = NotImplemented
            self.last_logoff = NotImplemented
            self.last_logon = NotImplemented
            self.last_seen_online = NotImplemented
            self.game = NotImplemented
            self.state = NotImplemented
            self.flags = NotImplemented
            self.privacy_state = NotImplemented
            self._is_commentable = NotImplemented
            self._setup_profile = NotImplemented
            self._level = data["level"]

        def __repr__(self) -> str:
            attrs = ("name", "id", "type", "universe", "instance")
            resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
            return f"<{self.__class__.__name__} {' '.join(resolved)}>"

        setattr(cls, "__init__", __init__)
        setattr(cls, "__repr__", __repr__)

        def not_implemented(function: str) -> None:
            @functools.wraps(getattr(cls, function))
            async def wrapped(*_, **__) -> None:
                raise NotImplementedError(
                    f"Accounts without an API key cannot use User.{function}, this is a Steam limitation not a library "
                    f"limitation, sorry."
                )

            setattr(cls, function, wrapped)

        not_implemented("friends")
        not_implemented("badges")
        not_implemented("is_commentable")
        not_implemented("has_setup_profile")
        not_implemented("is_private")
        not_implemented("is_banned")


@runtime_checkable
class Messageable(Protocol[M_co]):
    """An ABC that details the common operations on a Steam message.
    The following classes implement this ABC:

        - :class:`~steam.User`
        - :class:`~steam.ClanChannel`
        - :class:`~steam.GroupChannel`
        - :class:`~steam.DMChannel`
    """

    __slots__ = ()

    @abc.abstractmethod
    def _message_func(self, content: str) -> Coroutine[Any, Any, M_co]:
        raise NotImplementedError

    @abc.abstractmethod
    def _image_func(self, image: Image) -> Coroutine[Any, Any, None]:
        raise NotImplementedError

    async def send(self, content: Any = None, image: Image | None = None) -> M_co | None:
        """Send a message to a certain destination.

        Parameters
        ----------
        content
            The content of the message to send.
        image
            The image to send to the user.

        Note
        ----
        Anything as passed to ``content`` is implicitly cast to a :class:`str`.

        Raises
        ------
        :exc:`~steam.HTTPException`
            Sending the message failed.
        :exc:`~steam.Forbidden`
            You do not have permission to send the message.

        Returns
        -------
        The sent message, only applicable if ``content`` is passed.
        """
        message = None
        if content is not None:
            message = await self._message_func(str(content))
        if image is not None:
            await self._image_func(image)

        return message


@attr.dataclass(slots=True)
class Channel(Messageable[M_co]):
    _state: ConnectionState
    clan: Clan | None = None
    group: Group | None = None

    @abc.abstractmethod
    def history(
        self,
        limit: int | None = 100,
        before: datetime | None = None,
        after: datetime | None = None,
    ) -> AsyncIterator[M_co]:
        """An :class:`~steam.iterators.AsyncIterator` for accessing a channel's :class:`steam.Message`\\s.

        Examples
        --------

        Usage:

        .. code-block:: python3

            async for message in channel.history(limit=10):
                print("Author:", message.author, "Said:", message.content)

        Flattening into a list:

        .. code-block:: python3

            messages = await channel.history(limit=50).flatten()
            # messages is now a list of Message

        All parameters are optional.

        Parameters
        ----------
        limit
            The maximum number of messages to search through. Setting this to ``None`` will fetch all of the channel's
            messages, but this will be a very slow operation.
        before
            A time to search for messages before.
        after
            A time to search for messages after.

        Yields
        ------
        :class:`~steam.Message`
        """
        raise NotImplementedError


def _clean_up_content(content: str) -> str:  # steam does weird stuff with content
    return content.replace(r"\[", "[").replace("\\\\", "\\")


STEAM_EPOCH = datetime(2005, 1, 1)


class Message:
    """Represents a message from a :class:`~steam.User`. This is a base class from which all messages inherit.

    The following classes implement this:

        - :class:`~steam.UserMessage`
        - :class:`~steam.GroupMessage`
        - :class:`~steam.ClanMessage`

    Attributes
    ----------
    channel
        The channel the message was sent in.
    content
        The message's content.

        Note
        ----
        This is **not** what you will see in the steam client see :attr:`clean_content` for that.

    clean_content
        The message's clean content without bbcode.
    author
        The message's author.
    created_at
        The time the message was sent at.
    group
        The group the message was sent in. Will be ``None`` if the message wasn't sent in a :class:`~steam.Group`.
    clan
        The clan the message was sent in. Will be ``None`` if the message wasn't sent in a :class:`~steam.Clan`.
    """

    __slots__ = (
        "author",
        "content",
        "channel",
        "clean_content",
        "created_at",
        "ordinal",
        "group",
        "clan",
        "mentions",
        "_id_cs",
        "_state",
    )
    author: Authors
    created_at: datetime

    def __init__(self, channel: Channel[Message], proto: Any):
        self._state: ConnectionState = channel._state
        self.channel = channel
        self.group: Group | None = channel.group
        self.clan: Clan | None = channel.clan
        self.content: str = _clean_up_content(proto.message)
        self.ordinal: int = proto.ordinal
        self.clean_content: str = getattr(proto, "message_no_bbcode", None) or self.content
        self.mentions: Mentions | None = getattr(proto, "mentions", None)
        # self.reactions: list["Emoticon"] = []

    def __repr__(self) -> str:
        attrs = ("author", "id", "channel")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"

    @cached_slot_property
    def id(self) -> int:
        """A unique identifier for every message sent in a channel."""
        # a u64 "snowflake-esk" id measuring of the number of seconds passed since Steam's EPOCH and then the
        # "sequence"/ordinal of the message.
        return int(
            f"{int((self.created_at - STEAM_EPOCH).total_seconds()):032b}{self.ordinal:032b}",
            base=2,
        )



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: steammessages_econ.steamclient.proto
# plugin: python-betterproto
# Last updated 09/09/2021

from dataclasses import dataclass
from typing import List

import betterproto


@dataclass(eq=False, repr=False)
class GetTradeOfferAccessTokenRequest(betterproto.Message):
    generate_new_token: bool = betterproto.bool_field(1)


@dataclass(eq=False, repr=False)
class GetTradeOfferAccessTokenResponse(betterproto.Message):
    trade_offer_access_token: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class GetItemShopOverlayAuthUrlRequest(betterproto.Message):
    return_url: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class GetItemShopOverlayAuthUrlResponse(betterproto.Message):
    url: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class GetAssetClassInfoRequest(betterproto.Message):
    language: str = betterproto.string_field(1)
    appid: int = betterproto.uint32_field(2)
    classes: List["GetAssetClassInfoRequestClass"] = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class GetAssetClassInfoRequestClass(betterproto.Message):
    classid: int = betterproto.uint64_field(1)
    instanceid: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class ItemDescriptionLine(betterproto.Message):
    type: str = betterproto.string_field(1)
    value: str = betterproto.string_field(2)
    color: str = betterproto.string_field(3)
    label: str = betterproto.string_field(4)


@dataclass(eq=False, repr=False)
class ItemAction(betterproto.Message):
    link: str = betterproto.string_field(1)
    name: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class ItemDescription(betterproto.Message):
    appid: int = betterproto.int32_field(1)
    classid: int = betterproto.uint64_field(2)
    instanceid: int = betterproto.uint64_field(3)
    currency: bool = betterproto.bool_field(4)
    background_color: str = betterproto.string_field(5)
    icon_url: str = betterproto.string_field(6)
    icon_url_large: str = betterproto.string_field(7)
    descriptions: List["ItemDescriptionLine"] = betterproto.message_field(8)
    tradable: bool = betterproto.bool_field(9)
    actions: List["ItemAction"] = betterproto.message_field(10)
    owner_descriptions: List["ItemDescriptionLine"] = betterproto.message_field(11)
    owner_actions: List["ItemAction"] = betterproto.message_field(12)
    fraudwarnings: List[str] = betterproto.string_field(13)
    name: str = betterproto.string_field(14)
    name_color: str = betterproto.string_field(15)
    type: str = betterproto.string_field(16)
    market_name: str = betterproto.string_field(17)
    market_hash_name: str = betterproto.string_field(18)
    market_fee: str = betterproto.string_field(19)
    market_fee_app: int = betterproto.int32_field(28)
    contained_item: "ItemDescription" = betterproto.message_field(20)
    market_actions: List["ItemAction"] = betterproto.message_field(21)
    commodity: bool = betterproto.bool_field(22)
    market_tradable_restriction: int = betterproto.int32_field(23)
    market_marketable_restriction: int = betterproto.int32_field(24)
    marketable: bool = betterproto.bool_field(25)
    tags: List["ItemTag"] = betterproto.message_field(26)
    item_expiration: str = betterproto.string_field(27)
    market_buy_country_restriction: str = betterproto.string_field(30)
    market_sell_country_restriction: str = betterproto.string_field(31)


@dataclass(eq=False, repr=False)
class ItemTag(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    category: str = betterproto.string_field(2)
    internal_name: str = betterproto.string_field(3)
    localized_category_name: str = betterproto.string_field(4)
    localized_tag_name: str = betterproto.string_field(5)
    color: str = betterproto.string_field(6)


@dataclass(eq=False, repr=False)
class GetAssetClassInfoResponse(betterproto.Message):
    descriptions: List["ItemDescription"] = betterproto.message_field(1)



"""
The MIT License (MIT)

Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import asyncio
from collections.abc import Coroutine
from datetime import timedelta
from typing import TYPE_CHECKING, Any

from .abc import BaseUser, Messageable, UserDict
from .enums import TradeOfferState
from .errors import ClientException, ConfirmationError
from .models import URL
from .profile import OwnedProfileItems, ProfileItem

if TYPE_CHECKING:
    from .clan import Clan
    from .game import Game
    from .group import Group
    from .image import Image
    from .message import UserMessage
    from .state import ConnectionState
    from .trade import Inventory, TradeOffer

__all__ = (
    "User",
    "ClientUser",
)


class User(BaseUser, Messageable["UserMessage"]):
    """Represents a Steam user's account.

    .. container:: operations

        .. describe:: x == y

            Checks if two users are equal.

        .. describe:: str(x)

            Returns the user's name.

    Attributes
    ----------
    name
        The user's username.
    state
        The current persona state of the account (e.g. LookingToTrade).
    game
        The Game instance attached to the user. Is ``None`` if the user isn't in a game or one that is recognised by the
        api.
    avatar_url
        The avatar url of the user. Uses the large (184x184 px) image url.
    real_name
        The user's real name defined by them. Could be ``None``.
    primary_clan
        The user's primary clan.
    created_at
        The time at which the user's account was created. Could be ``None``.
    last_logon
        The last time the user logged into steam. This is only ``None`` if user hasn't been updated from the websocket.
    last_logoff
        The last time the user logged off from steam. Could be ``None`` (e.g. if they are currently online).
    last_seen_online
        The last time the user could be seen online. This is only ``None`` if user hasn't been updated from the
        websocket.
    country
        The country code of the account. Could be ``None``.
    flags
        The persona state flags of the account.
    """

    __slots__ = ()

    async def add(self) -> None:
        """Sends a friend invite to the user to your friends list."""
        await self._state.http.add_user(self.id64)

    async def remove(self) -> None:
        """Remove the user from your friends list."""
        await self._state.http.remove_user(self.id64)
        try:
            self._state.client.user.friends.remove(self)
        except ValueError:
            pass

    async def cancel_invite(self) -> None:
        """Cancels an invite sent to the user. This effectively does the same thing as :meth:`remove`."""
        await self._state.http.remove_user(self.id64)

    async def block(self) -> None:
        """Blocks the user."""
        await self._state.http.block_user(self.id64)

    async def unblock(self) -> None:
        """Unblocks the user."""
        await self._state.http.unblock_user(self.id64)

    async def escrow(self, token: str | None = None) -> timedelta | None:
        """Check how long any received items would take to arrive. ``None`` if the user has no escrow or has a
        private inventory.

        Parameters
        ----------
        token
            The user's trade offer token, not required if you are friends with the user.
        """
        resp = await self._state.http.get_user_escrow(self.id64, token)
        their_escrow = resp["response"].get("their_escrow")
        if their_escrow is None:  # private
            return None
        seconds = their_escrow["escrow_end_duration_seconds"]
        return timedelta(seconds=seconds) if seconds else None

    def _message_func(self, content: str) -> Coroutine[Any, Any, UserMessage]:
        return self._state.send_user_message(self.id64, content)

    def _image_func(self, image: Image) -> Coroutine[Any, Any, None]:
        return self._state.http.send_user_image(self.id64, image)

    async def send(
        self,
        content: Any = None,
        *,
        trade: TradeOffer | None = None,
        image: Image | None = None,
    ) -> UserMessage | None:
        """Send a message, trade or image to an :class:`User`.

        Parameters
        ----------
        content
            The message to send to the user.
        trade
            The trade offer to send to the user.

            Note
            ----
            This will have its :attr:`~steam.TradeOffer.id` attribute updated after being sent.

        image
            The image to send to the user.

        Raises
        ------
        :exc:`~steam.HTTPException`
            Sending the message failed.
        :exc:`~steam.Forbidden`
            You do not have permission to send the message.

        Returns
        -------
        The sent message only applicable if ``content`` is passed.
        """

        message = await super().send(content, image)
        if trade is not None:
            to_send = [item.to_dict() for item in trade.items_to_send]
            to_receive = [item.to_dict() for item in trade.items_to_receive]
            resp = await self._state.http.send_trade_offer(self, to_send, to_receive, trade.token, trade.message or "")
            trade._has_been_sent = True
            needs_confirmation = resp.get("needs_mobile_confirmation", False)
            trade._update_from_send(self._state, resp, self, active=not needs_confirmation)
            if needs_confirmation:
                for tries in range(5):
                    try:
                        await trade.confirm()
                    except ConfirmationError:
                        break
                    except ClientException:
                        await asyncio.sleep(tries * 2)
                trade.state = TradeOfferState.Active

            # make sure the trade is updated before this function returns
            self._state._trades[trade.id] = trade
            self._state._trades_to_watch.add(trade.id)
            await self._state.wait_for_trade(trade.id)
            self._state.dispatch("trade_send", trade)

        return message

    async def invite_to_group(self, group: Group) -> None:
        """Invites the user to a :class:`Group`.

        Parameters
        -----------
        group
            The group to invite the user to.
        """
        await self._state.invite_user_to_group(self.id64, group.id)

    async def invite_to_clan(self, clan: Clan) -> None:
        """Invites the user to a :class:`Clan`.

        Parameters
        -----------
        clan
            The clan to invite the user to.
        """
        await self._state.http.invite_user_to_clan(self.id64, clan.id64)

    async def owns(self, game: Game) -> bool:
        """Whether or not the game is owned by this user.

        Parameters
        ----------
        game
            The game you want to check the ownership of.
        """
        return self.id64 in await self._state.fetch_friends_who_own(game.id)

    def is_friend(self) -> bool:
        """Whether or not the user is in the :class:`ClientUser`'s friends."""
        return self in self._state.client.user.friends


class ClientUser(BaseUser):
    """Represents your account.

    .. container:: operations

        .. describe:: x == y

            Checks if two users are equal.

        .. describe:: str(x)

            Returns the user's name.

    Attributes
    ----------
    name
        The user's username.
    friends
        A list of the :class:`ClientUser`'s friends.
    state
        The current persona state of the account (e.g. LookingToTrade).
    game
        The Game instance attached to the user. Is ``None`` if the user isn't in a game or one that is recognised by
        the api.
    avatar_url
        The avatar url of the user. Uses the large (184x184 px) image url.
    real_name
        The user's real name defined by them. Could be ``None``.
    primary_clan
        The user's primary clan. Could be ``None``
    created_at
        The time at which the user's account was created. Could be ``None``.
    last_logon
        The last time the user logged into steam. This is only ``None`` if user hasn't been updated from the websocket.
    last_logoff
        The last time the user logged off from steam. Could be ``None`` (e.g. if they are currently online).
    last_seen_online
        The last time the user could be seen online. This is only ``None`` if user hasn't been updated from the
        websocket.
    country
        The country code of the account. Could be ``None``.
    flags
        The persona state flags of the account.
    """

    # TODO more stuff to add https://github.com/DoctorMcKay/node-steamcommunity/blob/master/components/profile.js

    __slots__ = ("friends", "_inventory_func")

    def __init__(self, state: ConnectionState, data: UserDict):
        super().__init__(state, data)
        self.friends: list[User] = []
        self._inventory_func = BaseUser.inventory

    async def inventory(self, game: Game) -> Inventory:
        return await self._inventory_func(self, game)

    async def setup_profile(self) -> None:
        """Set up your profile if possible."""
        if self.has_setup_profile():
            return

        params = {"welcomed": 1}
        await self._state.http.get(URL.COMMUNITY / "my/edit", params=params)

    async def clear_nicks(self) -> None:
        """Clears the client user's nickname/alias history."""
        await self._state.http.clear_nickname_history()

    async def profile_items(self) -> OwnedProfileItems:
        """Fetch all of the client user's profile items."""
        items = await self._state.fetch_profile_items()
        return OwnedProfileItems(
            backgrounds=[
                ProfileItem(self._state, background, um_name="ProfileBackground")
                for background in items.profile_backgrounds
            ],
            mini_profile_backgrounds=[
                ProfileItem(self._state, mini_profile_background, um_name="MiniProfileBackground")
                for mini_profile_background in items.mini_profile_backgrounds
            ],
            avatar_frames=[
                ProfileItem(self._state, avatar_frame, um_name="AvatarFrame") for avatar_frame in items.avatar_frames
            ],
            animated_avatars=[
                ProfileItem(self._state, animated_avatar, um_name="AnimatedAvatar")
                for animated_avatar in items.animated_avatars
            ],
            modifiers=[ProfileItem(self._state, modifier) for modifier in items.profile_modifiers],
        )

    async def edit(
        self,
        *,
        name: str | None = None,
        real_name: str | None = None,
        url: str | None = None,
        summary: str | None = None,
        country: str | None = None,
        state: str | None = None,
        city: str | None = None,
        avatar: Image | None = None,
    ) -> None:
        """Edit the client user's profile. Any values that aren't set will use their defaults.

        Parameters
        ----------
        name
            The new name you wish to go by.
        real_name
            The real name you wish to go by.
        url
            The custom url ending/path you wish to use.
        summary
            The summary/description you wish to use.
        country
            The country you want to be from.
        state
            The state you want to be from.
        city
            The city you want to be from.
        avatar
            The avatar you wish to use.

            Note
            ----
            This needs to be at least 184px x 184px.

        Raises
        -------
        :exc:`~steam.HTTPException`
            Editing your profile failed.
        """
        await self._state.http.edit_profile(name, real_name, url, summary, country, state, city, avatar)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: steammessages_publishedfile.steamclient.proto
# plugin: python-betterproto
# Last updated 09/09/2021

from dataclasses import dataclass
from typing import List

import betterproto

from .player import EBanContentCheckResult


class EPublishedFileRevision(betterproto.Enum):
    Default = 0
    Latest = 1
    ApprovedSnapshot = 2
    ApprovedSnapshotChina = 3
    RejectedSnapshot = 4
    RejectedSnapshotChina = 5


class EPublishedFileForSaleStatus(betterproto.Enum):
    NotForSale = 0
    PendingApproval = 1
    ApprovedForSale = 2
    RejectedForSale = 3
    NoLongerForSale = 4
    TentativeApproval = 5


@dataclass(eq=False, repr=False)
class SubscribeRequest(betterproto.Message):
    publishedfileid: int = betterproto.uint64_field(1)
    list_type: int = betterproto.uint32_field(2)
    appid: int = betterproto.int32_field(3)
    notify_client: bool = betterproto.bool_field(4)


@dataclass(eq=False, repr=False)
class SubscribeResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class UnsubscribeRequest(betterproto.Message):
    publishedfileid: int = betterproto.uint64_field(1)
    list_type: int = betterproto.uint32_field(2)
    appid: int = betterproto.int32_field(3)
    notify_client: bool = betterproto.bool_field(4)


@dataclass(eq=False, repr=False)
class UnsubscribeResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class CanSubscribeRequest(betterproto.Message):
    publishedfileid: int = betterproto.uint64_field(1)


@dataclass(eq=False, repr=False)
class CanSubscribeResponse(betterproto.Message):
    can_subscribe: bool = betterproto.bool_field(1)


@dataclass(eq=False, repr=False)
class PublishRequest(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    consumer_appid: int = betterproto.uint32_field(2)
    cloudfilename: str = betterproto.string_field(3)
    preview_cloudfilename: str = betterproto.string_field(4)
    title: str = betterproto.string_field(5)
    file_description: str = betterproto.string_field(6)
    file_type: int = betterproto.uint32_field(7)
    consumer_shortcut_name: str = betterproto.string_field(8)
    youtube_username: str = betterproto.string_field(9)
    youtube_videoid: str = betterproto.string_field(10)
    visibility: int = betterproto.uint32_field(11)
    redirect_uri: str = betterproto.string_field(12)
    tags: List[str] = betterproto.string_field(13)
    collection_type: str = betterproto.string_field(14)
    game_type: str = betterproto.string_field(15)
    url: str = betterproto.string_field(16)


@dataclass(eq=False, repr=False)
class PublishResponse(betterproto.Message):
    publishedfileid: int = betterproto.uint64_field(1)
    redirect_uri: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class GetDetailsRequest(betterproto.Message):
    publishedfileids: List[int] = betterproto.fixed64_field(1)
    includetags: bool = betterproto.bool_field(2)
    includeadditionalpreviews: bool = betterproto.bool_field(3)
    includechildren: bool = betterproto.bool_field(4)
    includekvtags: bool = betterproto.bool_field(5)
    includevotes: bool = betterproto.bool_field(6)
    short_description: bool = betterproto.bool_field(8)
    includeforsaledata: bool = betterproto.bool_field(10)
    includemetadata: bool = betterproto.bool_field(11)
    language: int = betterproto.int32_field(12)
    return_playtime_stats: int = betterproto.uint32_field(13)
    appid: int = betterproto.uint32_field(14)
    strip_description_bbcode: bool = betterproto.bool_field(15)
    desired_revision: "EPublishedFileRevision" = betterproto.enum_field(16)
    includereactions: bool = betterproto.bool_field(17)


@dataclass(eq=False, repr=False)
class PublishedFileDetails(betterproto.Message):
    result: int = betterproto.uint32_field(1)
    publishedfileid: int = betterproto.uint64_field(2)
    creator: int = betterproto.fixed64_field(3)
    creator_appid: int = betterproto.uint32_field(4)
    consumer_appid: int = betterproto.uint32_field(5)
    consumer_shortcutid: int = betterproto.uint32_field(6)
    filename: str = betterproto.string_field(7)
    file_size: int = betterproto.uint64_field(8)
    preview_file_size: int = betterproto.uint64_field(9)
    file_url: str = betterproto.string_field(10)
    preview_url: str = betterproto.string_field(11)
    youtubevideoid: str = betterproto.string_field(12)
    url: str = betterproto.string_field(13)
    hcontent_file: int = betterproto.fixed64_field(14)
    hcontent_preview: int = betterproto.fixed64_field(15)
    title: str = betterproto.string_field(16)
    file_description: str = betterproto.string_field(17)
    short_description: str = betterproto.string_field(18)
    time_created: int = betterproto.uint32_field(19)
    time_updated: int = betterproto.uint32_field(20)
    visibility: int = betterproto.uint32_field(21)
    flags: int = betterproto.uint32_field(22)
    workshop_file: bool = betterproto.bool_field(23)
    workshop_accepted: bool = betterproto.bool_field(24)
    show_subscribe_all: bool = betterproto.bool_field(25)
    num_comments_developer: int = betterproto.int32_field(26)
    num_comments_public: int = betterproto.int32_field(27)
    banned: bool = betterproto.bool_field(28)
    ban_reason: str = betterproto.string_field(29)
    banner: int = betterproto.fixed64_field(30)
    can_be_deleted: bool = betterproto.bool_field(31)
    incompatible: bool = betterproto.bool_field(32)
    app_name: str = betterproto.string_field(33)
    file_type: int = betterproto.uint32_field(34)
    can_subscribe: bool = betterproto.bool_field(35)
    subscriptions: int = betterproto.uint32_field(36)
    favorited: int = betterproto.uint32_field(37)
    followers: int = betterproto.uint32_field(38)
    lifetime_subscriptions: int = betterproto.uint32_field(39)
    lifetime_favorited: int = betterproto.uint32_field(40)
    lifetime_followers: int = betterproto.uint32_field(41)
    lifetime_playtime: int = betterproto.uint64_field(62)
    lifetime_playtime_sessions: int = betterproto.uint64_field(63)
    views: int = betterproto.uint32_field(42)
    image_width: int = betterproto.uint32_field(43)
    image_height: int = betterproto.uint32_field(44)
    image_url: str = betterproto.string_field(45)
    spoiler_tag: bool = betterproto.bool_field(46)
    shortcutid: int = betterproto.uint32_field(47)
    shortcutname: str = betterproto.string_field(48)
    num_children: int = betterproto.uint32_field(49)
    num_reports: int = betterproto.uint32_field(50)
    previews: List["PublishedFileDetailsPreview"] = betterproto.message_field(51)
    tags: List["PublishedFileDetailsTag"] = betterproto.message_field(52)
    children: List["PublishedFileDetailsChild"] = betterproto.message_field(53)
    kvtags: List["PublishedFileDetailsKvTag"] = betterproto.message_field(54)
    vote_data: "PublishedFileDetailsVoteData" = betterproto.message_field(55)
    playtime_stats: "PublishedFileDetailsPlaytimeStats" = betterproto.message_field(64)
    time_subscribed: int = betterproto.uint32_field(56)
    for_sale_data: "PublishedFileDetailsForSaleData" = betterproto.message_field(57)
    metadata: str = betterproto.string_field(58)
    language: int = betterproto.int32_field(61)
    maybe_inappropriate_sex: bool = betterproto.bool_field(65)
    maybe_inappropriate_violence: bool = betterproto.bool_field(66)
    revision_change_number: int = betterproto.uint64_field(67)
    revision: "EPublishedFileRevision" = betterproto.enum_field(68)
    available_revisions: List["EPublishedFileRevision"] = betterproto.enum_field(69)
    reactions: List["PublishedFileDetailsReaction"] = betterproto.message_field(70)
    ban_text_check_result: EBanContentCheckResult = betterproto.enum_field(71)


@dataclass(eq=False, repr=False)
class PublishedFileDetailsTag(betterproto.Message):
    tag: str = betterproto.string_field(1)
    adminonly: bool = betterproto.bool_field(2)
    display_name: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class PublishedFileDetailsPreview(betterproto.Message):
    previewid: int = betterproto.uint64_field(1)
    sortorder: int = betterproto.uint32_field(2)
    url: str = betterproto.string_field(3)
    size: int = betterproto.uint32_field(4)
    filename: str = betterproto.string_field(5)
    youtubevideoid: str = betterproto.string_field(6)
    preview_type: int = betterproto.uint32_field(7)
    external_reference: str = betterproto.string_field(8)


@dataclass(eq=False, repr=False)
class PublishedFileDetailsChild(betterproto.Message):
    publishedfileid: int = betterproto.uint64_field(1)
    sortorder: int = betterproto.uint32_field(2)
    file_type: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class PublishedFileDetailsKvTag(betterproto.Message):
    key: str = betterproto.string_field(1)
    value: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class PublishedFileDetailsVoteData(betterproto.Message):
    score: float = betterproto.float_field(1)
    votes_up: int = betterproto.uint32_field(2)
    votes_down: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class PublishedFileDetailsForSaleData(betterproto.Message):
    is_for_sale: bool = betterproto.bool_field(1)
    price_category: int = betterproto.uint32_field(2)
    estatus: "EPublishedFileForSaleStatus" = betterproto.enum_field(3)
    price_category_floor: int = betterproto.uint32_field(4)
    price_is_pay_what_you_want: bool = betterproto.bool_field(5)
    discount_percentage: int = betterproto.uint32_field(6)


@dataclass(eq=False, repr=False)
class PublishedFileDetailsPlaytimeStats(betterproto.Message):
    playtime_seconds: int = betterproto.uint64_field(1)
    num_sessions: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class PublishedFileDetailsReaction(betterproto.Message):
    reactionid: int = betterproto.uint32_field(1)
    count: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class GetDetailsResponse(betterproto.Message):
    publishedfiledetails: List["PublishedFileDetails"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class GetItemInfoRequest(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    last_time_updated: int = betterproto.uint32_field(2)
    workshop_items: List["GetItemInfoRequestWorkshopItem"] = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class GetItemInfoRequestWorkshopItem(betterproto.Message):
    published_file_id: int = betterproto.fixed64_field(1)
    time_updated: int = betterproto.uint32_field(2)
    desired_revision: "EPublishedFileRevision" = betterproto.enum_field(3)


@dataclass(eq=False, repr=False)
class GetItemInfoResponse(betterproto.Message):
    update_time: int = betterproto.uint32_field(1)
    workshop_items: List["GetItemInfoResponseWorkshopItemInfo"] = betterproto.message_field(2)
    private_items: List[int] = betterproto.fixed64_field(3)


@dataclass(eq=False, repr=False)
class GetItemInfoResponseWorkshopItemInfo(betterproto.Message):
    published_file_id: int = betterproto.fixed64_field(1)
    time_updated: int = betterproto.uint32_field(2)
    manifest_id: int = betterproto.fixed64_field(3)
    flags: int = betterproto.uint32_field(4)


@dataclass(eq=False, repr=False)
class GetUserFilesRequest(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)
    appid: int = betterproto.uint32_field(2)
    page: int = betterproto.uint32_field(4)
    numperpage: int = betterproto.uint32_field(5)
    type: str = betterproto.string_field(6)
    sortmethod: str = betterproto.string_field(7)
    privacy: int = betterproto.uint32_field(9)
    requiredtags: List[str] = betterproto.string_field(10)
    excludedtags: List[str] = betterproto.string_field(11)
    required_kv_tags: List["GetUserFilesRequestKvTag"] = betterproto.message_field(30)
    filetype: int = betterproto.uint32_field(14)
    creator_appid: int = betterproto.uint32_field(15)
    match_cloud_filename: str = betterproto.string_field(16)
    cache_max_age_seconds: int = betterproto.uint32_field(27)
    language: int = betterproto.int32_field(29)
    taggroups: List["GetUserFilesRequestTagGroup"] = betterproto.message_field(34)
    totalonly: bool = betterproto.bool_field(17)
    ids_only: bool = betterproto.bool_field(18)
    return_vote_data: bool = betterproto.bool_field(19)
    return_tags: bool = betterproto.bool_field(20)
    return_kv_tags: bool = betterproto.bool_field(21)
    return_previews: bool = betterproto.bool_field(22)
    return_children: bool = betterproto.bool_field(23)
    return_short_description: bool = betterproto.bool_field(24)
    return_for_sale_data: bool = betterproto.bool_field(26)
    return_metadata: bool = betterproto.bool_field(28)
    return_playtime_stats: int = betterproto.uint32_field(31)
    strip_description_bbcode: bool = betterproto.bool_field(32)
    return_reactions: bool = betterproto.bool_field(35)
    startindex_override: int = betterproto.uint32_field(25)
    desired_revision: "EPublishedFileRevision" = betterproto.enum_field(33)


@dataclass(eq=False, repr=False)
class GetUserFilesRequestKvTag(betterproto.Message):
    key: str = betterproto.string_field(1)
    value: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class GetUserFilesRequestTagGroup(betterproto.Message):
    tags: List[str] = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class GetUserFilesResponse(betterproto.Message):
    total: int = betterproto.uint32_field(1)
    startindex: int = betterproto.uint32_field(2)
    publishedfiledetails: List["PublishedFileDetails"] = betterproto.message_field(3)
    apps: List["GetUserFilesResponseApp"] = betterproto.message_field(4)


@dataclass(eq=False, repr=False)
class GetUserFilesResponseApp(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    name: str = betterproto.string_field(2)
    shortcutid: int = betterproto.uint32_field(3)
    private: bool = betterproto.bool_field(4)


@dataclass(eq=False, repr=False)
class AreFilesInSubscriptionListRequest(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    publishedfileids: List[int] = betterproto.fixed64_field(2)
    listtype: int = betterproto.uint32_field(3)
    filetype: int = betterproto.uint32_field(4)
    workshopfiletype: int = betterproto.uint32_field(5)


@dataclass(eq=False, repr=False)
class AreFilesInSubscriptionListResponse(betterproto.Message):
    files: List["AreFilesInSubscriptionListResponseInList"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class AreFilesInSubscriptionListResponseInList(betterproto.Message):
    publishedfileid: int = betterproto.fixed64_field(1)
    inlist: bool = betterproto.bool_field(2)


@dataclass(eq=False, repr=False)
class UpdateRequest(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    publishedfileid: int = betterproto.fixed64_field(2)
    title: str = betterproto.string_field(3)
    file_description: str = betterproto.string_field(4)
    visibility: int = betterproto.uint32_field(5)
    tags: List[str] = betterproto.string_field(6)
    filename: str = betterproto.string_field(7)
    preview_filename: str = betterproto.string_field(8)
    image_width: int = betterproto.uint32_field(15)
    image_height: int = betterproto.uint32_field(16)


@dataclass(eq=False, repr=False)
class UpdateResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class GetChangeHistoryEntryRequest(betterproto.Message):
    publishedfileid: int = betterproto.fixed64_field(1)
    timestamp: int = betterproto.uint32_field(2)
    language: int = betterproto.int32_field(3)


@dataclass(eq=False, repr=False)
class GetChangeHistoryEntryResponse(betterproto.Message):
    change_description: str = betterproto.string_field(1)
    language: int = betterproto.int32_field(2)


@dataclass(eq=False, repr=False)
class GetChangeHistoryRequest(betterproto.Message):
    publishedfileid: int = betterproto.fixed64_field(1)
    total_only: bool = betterproto.bool_field(2)
    startindex: int = betterproto.uint32_field(3)
    count: int = betterproto.uint32_field(4)
    language: int = betterproto.int32_field(5)


@dataclass(eq=False, repr=False)
class GetChangeHistoryResponse(betterproto.Message):
    changes: List["GetChangeHistoryResponseChangeLog"] = betterproto.message_field(1)
    total: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class GetChangeHistoryResponseChangeLog(betterproto.Message):
    timestamp: int = betterproto.uint32_field(1)
    change_description: str = betterproto.string_field(2)
    language: int = betterproto.int32_field(3)


@dataclass(eq=False, repr=False)
class RefreshVotingQueueRequest(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    matching_file_type: int = betterproto.uint32_field(2)
    tags: List[str] = betterproto.string_field(3)
    match_all_tags: bool = betterproto.bool_field(4)
    excluded_tags: List[str] = betterproto.string_field(5)
    desired_queue_size: int = betterproto.uint32_field(6)
    desired_revision: "EPublishedFileRevision" = betterproto.enum_field(8)


@dataclass(eq=False, repr=False)
class RefreshVotingQueueResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class QueryFilesRequest(betterproto.Message):
    query_type: int = betterproto.uint32_field(1)
    page: int = betterproto.uint32_field(2)
    cursor: str = betterproto.string_field(39)
    numperpage: int = betterproto.uint32_field(3)
    creator_appid: int = betterproto.uint32_field(4)
    appid: int = betterproto.uint32_field(5)
    requiredtags: List[str] = betterproto.string_field(6)
    excludedtags: List[str] = betterproto.string_field(7)
    match_all_tags: bool = betterproto.bool_field(8)
    required_flags: List[str] = betterproto.string_field(9)
    omitted_flags: List[str] = betterproto.string_field(10)
    search_text: str = betterproto.string_field(11)
    filetype: int = betterproto.uint32_field(12)
    child_publishedfileid: int = betterproto.fixed64_field(13)
    days: int = betterproto.uint32_field(14)
    include_recent_votes_only: bool = betterproto.bool_field(15)
    cache_max_age_seconds: int = betterproto.uint32_field(31)
    language: int = betterproto.int32_field(33)
    required_kv_tags: List["QueryFilesRequestKvTag"] = betterproto.message_field(34)
    taggroups: List["QueryFilesRequestTagGroup"] = betterproto.message_field(42)
    totalonly: bool = betterproto.bool_field(16)
    ids_only: bool = betterproto.bool_field(35)
    return_vote_data: bool = betterproto.bool_field(17)
    return_tags: bool = betterproto.bool_field(18)
    return_kv_tags: bool = betterproto.bool_field(19)
    return_previews: bool = betterproto.bool_field(20)
    return_children: bool = betterproto.bool_field(21)
    return_short_description: bool = betterproto.bool_field(22)
    return_for_sale_data: bool = betterproto.bool_field(30)
    return_metadata: bool = betterproto.bool_field(32)
    return_playtime_stats: int = betterproto.uint32_field(36)
    return_details: bool = betterproto.bool_field(37)
    strip_description_bbcode: bool = betterproto.bool_field(38)
    desired_revision: "EPublishedFileRevision" = betterproto.enum_field(40)
    return_reactions: bool = betterproto.bool_field(43)


@dataclass(eq=False, repr=False)
class QueryFilesRequestKvTag(betterproto.Message):
    key: str = betterproto.string_field(1)
    value: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class QueryFilesRequestTagGroup(betterproto.Message):
    tags: List[str] = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class QueryFilesResponse(betterproto.Message):
    total: int = betterproto.uint32_field(1)
    publishedfiledetails: List["PublishedFileDetails"] = betterproto.message_field(2)
    next_cursor: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class AddAppRelationshipRequest(betterproto.Message):
    publishedfileid: int = betterproto.uint64_field(1)
    appid: int = betterproto.uint32_field(2)
    relationship: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class AddAppRelationshipResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class RemoveAppRelationshipRequest(betterproto.Message):
    publishedfileid: int = betterproto.uint64_field(1)
    appid: int = betterproto.uint32_field(2)
    relationship: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class RemoveAppRelationshipResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class GetAppRelationshipsRequest(betterproto.Message):
    publishedfileid: int = betterproto.uint64_field(1)


@dataclass(eq=False, repr=False)
class GetAppRelationshipsResponse(betterproto.Message):
    app_relationships: List["GetAppRelationshipsResponseAppRelationship"] = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class GetAppRelationshipsResponseAppRelationship(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    relationship: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class StartPlaytimeTrackingRequest(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    publishedfileids: List[int] = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class StartPlaytimeTrackingResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class StopPlaytimeTrackingRequest(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    publishedfileids: List[int] = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class StopPlaytimeTrackingResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class StopPlaytimeTrackingForAllAppItemsRequest(betterproto.Message):
    appid: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class StopPlaytimeTrackingForAllAppItemsResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class SetPlaytimeForControllerConfigsRequest(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    controller_config_usage: List[
        "SetPlaytimeForControllerConfigsRequestControllerConfigUsage"
    ] = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class SetPlaytimeForControllerConfigsRequestControllerConfigUsage(betterproto.Message):
    publishedfileid: int = betterproto.uint64_field(1)
    seconds_active: float = betterproto.float_field(2)


@dataclass(eq=False, repr=False)
class SetPlaytimeForControllerConfigsResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class AddChildRequest(betterproto.Message):
    publishedfileid: int = betterproto.uint64_field(1)
    child_publishedfileid: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class AddChildResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class RemoveChildRequest(betterproto.Message):
    publishedfileid: int = betterproto.uint64_field(1)
    child_publishedfileid: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class RemoveChildResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class GetUserVoteSummaryRequest(betterproto.Message):
    publishedfileids: List[int] = betterproto.fixed64_field(1)


@dataclass(eq=False, repr=False)
class GetUserVoteSummaryResponse(betterproto.Message):
    summaries: List["GetUserVoteSummaryResponseVoteSummary"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class GetUserVoteSummaryResponseVoteSummary(betterproto.Message):
    publishedfileid: int = betterproto.fixed64_field(1)
    vote_for: bool = betterproto.bool_field(2)
    vote_against: bool = betterproto.bool_field(3)
    reported: bool = betterproto.bool_field(4)


@dataclass(eq=False, repr=False)
class GetItemChangesRequest(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    last_time_updated: int = betterproto.uint32_field(2)
    num_items_max: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class GetItemChangesResponse(betterproto.Message):
    update_time: int = betterproto.uint32_field(1)
    workshop_items: List["GetItemChangesResponseWorkshopItemInfo"] = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class GetItemChangesResponseWorkshopItemInfo(betterproto.Message):
    published_file_id: int = betterproto.fixed64_field(1)
    time_updated: int = betterproto.uint32_field(2)
    manifest_id: int = betterproto.fixed64_field(3)


@dataclass(eq=False, repr=False)
class FileSubscribedNotification(betterproto.Message):
    published_file_id: int = betterproto.fixed64_field(1)
    app_id: int = betterproto.uint32_field(2)
    file_hcontent: int = betterproto.fixed64_field(3)
    file_size: int = betterproto.uint32_field(4)
    rtime_subscribed: int = betterproto.uint32_field(5)
    is_depot_content: bool = betterproto.bool_field(6)
    rtime_updated: int = betterproto.uint32_field(7)
    revisions: List["FileSubscribedNotificationRevisionData"] = betterproto.message_field(8)


@dataclass(eq=False, repr=False)
class FileSubscribedNotificationRevisionData(betterproto.Message):
    revision: "EPublishedFileRevision" = betterproto.enum_field(1)
    file_hcontent: int = betterproto.fixed64_field(2)
    rtime_updated: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class FileUnsubscribedNotification(betterproto.Message):
    published_file_id: int = betterproto.fixed64_field(1)
    app_id: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class FileDeletedClientNotification(betterproto.Message):
    published_file_id: int = betterproto.fixed64_field(1)
    app_id: int = betterproto.uint32_field(2)



from dataclasses import dataclass
from typing import List

import betterproto

from .base import CClanEventData


@dataclass(eq=False, repr=False)
class RespondToClanInviteRequest(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)
    accept: bool = betterproto.bool_field(2)


@dataclass(eq=False, repr=False)
class RespondToClanInviteResponse(betterproto.Message):
    pass


# not public and also non-functional atm
@dataclass(eq=False, repr=False)
class GetAdjacentPartnerEventsRequest(betterproto.Message):
    steam_id: int = betterproto.fixed64_field(1)  # might be account id
    announcement_id: int = betterproto.fixed64_field(2)
    # count_later: int = betterproto.uint32_field(2)  # might allow announcement_id
    # might allow the limit


@dataclass(eq=False, repr=False)
class GetAdjacentPartnerEventsResponse(betterproto.Message):
    events: List[CClanEventData] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class GetEventDetailsRequest(betterproto.Message):
    event_ids: List[int] = betterproto.int64_field(1)
    clan_ids: List[int] = betterproto.fixed32_field(2)


@dataclass(eq=False, repr=False)
class GetEventDetailsResponse(betterproto.Message):
    events: List[CClanEventData] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class GetSinglePartnerEventRequest(betterproto.Message):
    clan_id: int = betterproto.fixed32_field(1)
    announcement_id: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class GetSinglePartnerEventResponse(betterproto.Message):
    events: List[CClanEventData] = betterproto.message_field(1)



from typing import NamedTuple

from typing_extensions import Literal

__all__ = (
    "__title__",
    "__author__",
    "__license__",
    "__version__",
    "version_info",
)


class VersionInfo(NamedTuple):
    major: int
    minor: int
    micro: int
    releaselevel: Literal["alpha", "beta", "candidate", "final"]


__title__ = "steam"
__author__ = "Gobot1234"
__license__ = "MIT"
__version__ = "0.8.5"
version_info = VersionInfo(major=0, minor=8, micro=5, releaselevel="final")



from __future__ import annotations

import asyncio
from typing import Any

from typing_extensions import ClassVar, Final

from ...abc import BaseUser
from ...client import Client as Client_
from ...game import Game
from ...protobufs import GCMsg, GCMsgProto
from ...trade import Inventory
from .state import GCState

__all__ = ("Client",)


class Client(Client_):
    _connection: GCState
    _GAME: Final[Game] = Any
    _GC_HEART_BEAT: ClassVar = 30.0

    def __init__(self, **options: Any):
        game = options.pop("game", None)
        if game is not None:  # don't let them overwrite the main game
            try:
                options["games"].append(game)
            except (TypeError, KeyError):
                options["games"] = [game]
        options["game"] = self._GAME
        self._original_games: list[Game] | None = options.get("games")
        super().__init__(**options)

    # things to override
    def _get_gc_message(self) -> GCMsgProto[Any] | GCMsg[Any]:
        raise NotImplementedError()

    def _get_state(self, **options: Any) -> None:
        raise NotImplementedError("cannot instantiate Client without a state")

    async def connect(self) -> None:
        if self._get_gc_message():

            async def ping_gc() -> None:
                await self.wait_until_ready()
                while not self.is_closed():
                    await self.ws.send_gc_message(self._get_gc_message())
                    await asyncio.sleep(self._GC_HEART_BEAT)

            await asyncio.gather(
                super().connect(),
                ping_gc(),
            )
        else:
            await super().connect()

    async def _handle_ready(self_) -> None:
        async def inventory(self: Any, game: Game) -> Inventory:
            return (
                self._connection.backpack
                if game == self.__class__._GAME and self_._connection._gc_ready.is_set()
                else await BaseUser.inventory(self, game)
            )

        self_._connection._unpatched_inventory = self_.user._inventory_func
        self_.user._inventory_func = inventory

        await super()._handle_ready()

    async def wait_for_gc_ready(self) -> None:
        await self._connection._gc_ready.wait()

    async def buy_item(self, def_id: int, price: int, def_ids: list[int], prices: int) -> None:
        ...



"""
The MIT License (MIT)

Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from . import utils
from .abc import SteamID
from .channel import GroupChannel
from .enums import Type
from .role import Role

if TYPE_CHECKING:
    from .protobufs import chat
    from .state import ConnectionState
    from .user import User


__all__ = ("Group",)


class Group(SteamID):
    """Represents a Steam group.

    Attributes
    ----------
    name
        The name of the group, could be ``None``.
    owner
        The owner of the group.
    members
        A list of the group's members.
    active_member_count
        The group's active member count.
    roles
        A list of the group's roles.
    default_role
        The group's default role.
    default_channel
        The group's default channel.
    """

    __slots__ = (
        "owner",
        "name",
        "members",
        "roles",
        "active_member_count",
        "default_role",
        "default_channel",
        "_channels",
        "_top_members_id64s",
        "_state",
    )

    owner: User | SteamID
    members: list[User]
    name: str | None
    active_member_count: int
    roles: list[Role]
    default_role: Role | None
    _channels: dict[int, GroupChannel]
    _top_members_id64s: list[int]

    # TODO
    # Group.ban
    # Group.kick
    # Group.create_channel

    def __init__(self, state: ConnectionState, id: int):
        super().__init__(id, type=Type.Chat)
        self._state = state

    @classmethod
    async def _from_proto(
        cls, state: ConnectionState, proto: chat.GetChatRoomGroupSummaryResponse | chat.GroupHeaderState
    ) -> Group:
        self = cls(state, proto.chat_group_id)
        self.owner = await self._state._maybe_user(utils.make_id64(proto.accountid_owner))
        self._top_members_id64s = proto.top_members
        self.name = proto.chat_group_name or None

        self.active_member_count = proto.active_member_count
        self.roles = []
        for role in await self._state.fetch_group_roles(self.id):
            for permissions in proto.role_actions:
                if permissions.role_id == role.role_id:
                    self.roles.append(Role(self._state, self, role, permissions))

        self.default_role = utils.get(self.roles, id=proto.default_role_id)
        self._channels = {
            channel.chat_id: GroupChannel(state=self._state, group=self, proto=channel) for channel in proto.chat_rooms
        }
        self.default_channel = self._channels.get(proto.default_chat_id)
        return self

    def __repr__(self) -> str:
        attrs = ("name", "id", "owner")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<Group {' '.join(resolved)}>"

    def __str__(self) -> str:
        return self.name or ""

    def _update(self, proto: chat.ChatRoomGroupRoomsChangeNotification) -> None:
        for channel in proto.chat_rooms:
            try:
                new_channel = self._channels[channel.chat_id]
            except KeyError:
                new_channel = GroupChannel(state=self._state, group=self, proto=channel)
                self._channels[new_channel.id] = new_channel
            else:
                new_channel._update(channel)
        self.default_channel = self._channels[int(proto.default_chat_id)]

    @property
    def channels(self) -> list[GroupChannel]:
        """A list of the group's channels."""
        return list(self._channels.values())

    @property
    def top_members(self) -> list[User | None]:
        """A list of the group's top members according to Steam."""
        return [self._state.get_user(id64) for id64 in self._top_members_id64s]

    async def leave(self) -> None:
        """Leaves the group."""
        await self._state.leave_chat(self.id)

    async def invite(self, user: User) -> None:
        """Invites a :class:`~steam.User` to the group.

        Parameters
        -----------
        user
            The user to invite to the group.
        """
        await self._state.invite_user_to_group(user.id64, self.id)



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz
Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

Heavily inspired by
https://github.com/Rapptz/discord.py/blob/master/discord/ext/commands/core.py
"""

from __future__ import annotations

import asyncio
import functools
import inspect
from collections.abc import Coroutine, Iterable
from time import time
from typing import TYPE_CHECKING, Any, Callable, ForwardRef, Generic, TypeVar, Union, get_type_hints, overload

from typing_extensions import Literal, ParamSpec, Protocol, Self, TypeAlias, get_args, get_origin

from ...channel import DMChannel
from ...errors import ClientException
from ...utils import cached_property, maybe_coroutine
from . import converters
from .cooldown import BucketType, Cooldown
from .errors import (
    BadArgument,
    CheckFailure,
    CommandDisabled,
    DMChannelOnly,
    DuplicateKeywordArgument,
    MissingRequiredArgument,
    NotOwner,
    UnmatchedKeyValuePair,
)
from .utils import CaseInsensitiveDict

if TYPE_CHECKING:
    from .bot import Bot
    from .cog import Cog
    from .context import Context

__all__ = (
    "Command",
    "command",
    "Group",
    "group",
    "check",
    "is_owner",
    "cooldown",
)

CheckType: TypeAlias = "Callable[[Context], Union[bool, Coroutine[Any, Any, bool]]]"
MaybeCommand: TypeAlias = "Callable[..., Command[Any]] | CallbackType"
C = TypeVar("C", bound="Command[Any]")
G = TypeVar("G", bound="Group[Any]")
Err = TypeVar("Err", bound="Callable[[Context, Exception], Coroutine[Any, Any, None]]")
InvokeT = TypeVar("InvokeT", bound="Callable[[Context], Coroutine[Any, Any, None]]")
MC = TypeVar("MC", bound=MaybeCommand)
MCD = TypeVar("MCD", bound="CommandDeco[MaybeCommand] | MaybeCommand")
CallT = TypeVar("CallT", bound="CallbackType")
CHR = TypeVar("CHR", bound="CheckReturnType")

P = ParamSpec("P")


class CommandDeco(Protocol):
    def __call__(self, __command: MC) -> MC:
        ...


class CheckReturnType(CommandDeco):
    predicate: CheckType


class CallbackType(Protocol[P]):
    __commands_checks__: list[CheckType]
    __commands_cooldown__: list[Cooldown[Any]]
    __special_converters__: list[type[converters.Converter]]

    async def __call__(self, *args: P.args, **kwargs: P.kwargs) -> Any:
        ...


@converters.converter_for(bool)
def to_bool(argument: str) -> bool:
    lowered = argument.lower()
    if lowered in ("yes", "y", "true", "t", "1", "enable", "on"):
        return True
    elif lowered in ("no", "n", "false", "f", "0", "disable", "off"):
        return False
    raise BadArgument(f"{argument!r} is not a recognised boolean option")


class Command(Generic[P]):
    """A class to represent a command.

    Attributes
    ----------
    name
        The command's name.
    help
        The command's help docstring.
    checks
        A list of the command's checks.
    cooldown
        The command's cooldowns.
    special_converters
        A list of the command's special converters as registered by :meth:`steam.ext.commands.Converter.register`.
    enabled
        Whether or not the command is enabled.
    cog
        The command's cog.
    parent
        The command's parent.
    description
        The command's description.
    hidden
        Whether or not the command is hidden.
    aliases
        The command's aliases.
    params
        The command's parameters.
    """

    DECORATORS: set[str] = {  # used in Cog._inject to update any unbound methods
        "on_error",
        "_before_hook",
        "_after_hook",
    }

    def __new__(cls: type[C], *args: Any, **kwargs: Any) -> C:
        self = super().__new__(cls)
        self.__original_kwargs__: dict[str, Any] = kwargs.copy()
        return self

    def __init__(self, func: CallbackType[P], **kwargs: Any):
        name = kwargs.get("name") or func.__name__
        if not isinstance(name, str):
            raise TypeError("name must be a string.")

        self.name = name
        self.callback = func

        help_doc = kwargs.get("help")
        self.help: str | None = inspect.cleandoc(help_doc) if help_doc is not None else inspect.getdoc(func)

        try:
            checks = func.__commands_checks__
            checks.reverse()
        except AttributeError:
            checks = kwargs.get("checks", [])
        finally:
            self.checks: list[CheckReturnType] = checks

        try:
            cooldown = func.__commands_cooldown__
        except AttributeError:
            cooldown = kwargs.get("cooldown", [])
        finally:
            self.cooldown: list[Cooldown] = cooldown

        try:
            special_converters = func.__special_converters__
        except AttributeError:
            special_converters = kwargs.get("special_converters", [])
        finally:
            self.special_converters: list[type[converters.Converter]] = special_converters

        self.enabled: bool = kwargs.get("enabled", True)
        self.brief: str | None = kwargs.get("brief")
        self.usage: str | None = kwargs.get("usage")
        self.cog: Cog | Bot | None = kwargs.get("cog")
        self.parent: GroupMixin | None = kwargs.get("parent")
        self.description: str = inspect.cleandoc(kwargs.get("description", ""))
        self.hidden: bool = kwargs.get("hidden", False)
        self.aliases: Iterable[str] = kwargs.get("aliases", ())

        for alias in self.aliases:
            if not isinstance(alias, str):
                raise TypeError("A commands aliases should be an iterable only containing strings")

        self._before_hook = None
        self._after_hook = None

    def __str__(self) -> str:
        return self.qualified_name

    @property
    def callback(self) -> CallbackType[P]:
        """The internal callback the command holds."""
        return self._callback

    @callback.setter
    def callback(self, function: CallbackType[P]) -> None:
        if not inspect.iscoroutinefunction(function):
            raise TypeError(f"The callback for the command {function.__name__!r} must be a coroutine function.")

        function = function.__func__ if inspect.ismethod(function) else function  # HelpCommand.command_callback

        annotations = get_type_hints(function)
        for name, annotation in annotations.items():
            if get_origin(annotation) is converters.Greedy and isinstance(annotation.converter, ForwardRef):
                annotations[name] = converters.Greedy[eval(annotation.converter.__forward_code__, function.__globals__)]

        function.__annotations__ = annotations

        self.params: dict[str, inspect.Parameter] = dict(inspect.signature(function).parameters)
        if not self.params:
            raise ClientException(f'Callback for {self.name} command is missing a "ctx" parameter.') from None

        self.module = function.__module__
        self._callback = function

    @cached_property
    def clean_params(self) -> dict[str, inspect.Parameter]:
        """The command's parameters without ``"self"`` and ``"ctx"``."""
        params = self.params.copy()
        keys = list(params)
        if self.cog is not None:
            try:
                del params[keys.pop(0)]  # cog's "self" param
            except IndexError:
                raise ClientException(f'Callback for {self.name} command is missing a "self" parameter.') from None
        try:
            del params[keys.pop(0)]  # context param
        except IndexError:
            raise ClientException(f'Callback for {self.name} command is missing a "ctx" parameter.') from None
        return params

    @property
    def qualified_name(self) -> str:
        """The full name of the command, this takes into account subcommands etc."""
        return " ".join(c.name for c in reversed(self.parents))

    @property
    def parents(self) -> list[Self]:
        """The command's parents.

        Returns
        -------
        :class:`list`\\[:class:`Command`]
        """
        commands = []
        command = self
        while command is not None:
            if not isinstance(command, Command):
                break
            commands.append(command)
            command = command.parent

        return commands

    async def __call__(self, ctx: Context, *args: P.args, **kwargs: P.kwargs) -> None:
        """Calls the internal callback that the command holds.

        Note
        ----
        This bypasses all mechanisms -- including checks, converters, invoke hooks, cooldowns, etc. You must take
        care to pass the proper arguments (excluding the parameter "self" in a cog context) and types to this
        function.
        """
        if self.cog is not None:
            return await self.callback(self.cog, ctx, *args, **kwargs)
        else:
            return await self.callback(ctx, *args, **kwargs)

    @overload
    def error(self, coro: None = ...) -> Callable[[Err], Err]:
        ...

    @overload
    def error(self, coro: Err) -> Err:
        ...

    def error(self, coro: Err | None = None) -> Callable[[Err], Err] | Err:
        """|maybecallabledeco|
        Register a :term:`coroutine function` to handle a commands ``on_error`` functionality similarly to
        :meth:`steam.ext.commands.Bot.on_command_error`.

        Example:

        .. code-block:: python3

            @bot.command
            async def raise_an_error(ctx: commands.Context) -> None:
                raise Exception("oh no an error")


            @raise_an_error.error
            async def on_error(ctx: commands.Context, error: Exception) -> None:
                await ctx.send(f"{ctx.command.name} raised an exception {error!r}")
        """

        def decorator(coro: Err) -> Err:
            if not inspect.iscoroutinefunction(coro):
                raise TypeError(f"Error handler for {self.name} must be a coroutine function")
            self.on_error = coro
            return coro

        return decorator(coro) if coro is not None else decorator

    @overload
    def before_invoke(self, coro: None = ...) -> Callable[[InvokeT], InvokeT]:
        ...

    @overload
    def before_invoke(self, coro: InvokeT) -> InvokeT:
        ...

    def before_invoke(self, coro: InvokeT | None = None) -> Callable[[InvokeT], InvokeT] | InvokeT:
        """|maybecallabledeco|
        Register a :term:`coroutine function` to be run before any arguments are parsed.
        """

        def decorator(coro: InvokeT) -> InvokeT:
            if not inspect.iscoroutinefunction(coro):
                raise TypeError(f"Hook for {self.name} must be a coroutine function")
            self._before_hook = coro
            return coro

        return decorator(coro) if coro is not None else decorator

    @overload
    def after_invoke(self, coro: None = ...) -> Callable[[InvokeT], InvokeT]:
        ...

    @overload
    def after_invoke(self, coro: InvokeT) -> InvokeT:
        ...

    def after_invoke(self, coro: InvokeT | None = None) -> Callable[[InvokeT], InvokeT] | InvokeT:
        """|maybecallabledeco|
        Register a :term:`coroutine function` to be run after the command has been invoked.
        """

        def decorator(coro: InvokeT) -> InvokeT:
            if not inspect.iscoroutinefunction(coro):
                raise TypeError(f"Hook for {self.name} must be a coroutine function")
            self._after_hook = coro
            return coro

        return decorator(coro) if coro is not None else decorator

    async def invoke(self, ctx: Context) -> None:
        """Invoke the callback the command holds.

        Parameters
        ----------
        ctx
            The invocation context.
        """
        try:
            try:  # we mess with mentions in UserConverter.convert, so we need to copy them for later
                mentions = ctx.message.mentions.ids.copy()
            except AttributeError:
                pass
            if not self.enabled:
                raise CommandDisabled(self)
            for check in ctx.bot.checks:
                if not await maybe_coroutine(check, ctx):
                    raise CheckFailure("You failed to pass one of the checks for this command")
            for check in self.checks:
                if not await maybe_coroutine(check, ctx):
                    raise CheckFailure("You failed to pass one of the checks for this command")
            for cooldown in self.cooldown:
                cooldown(ctx)
            await self._parse_arguments(ctx)
            await self._call_before_invoke(ctx)
            await self(ctx, *ctx.args, **ctx.kwargs)
        except asyncio.CancelledError:
            return
        except Exception:
            ctx.command_failed = True
            raise
        finally:
            try:
                ctx.message.mentions.ids = mentions
            except (UnboundLocalError, AttributeError):
                pass
            await self._call_after_invoke(ctx)

    async def can_run(self, ctx: Context) -> bool:
        """Whether the command can be run.

        Parameters
        ----------
        ctx
            The invocation context.
        """
        if not self.enabled:
            return False
        for check in self.checks:
            if not await maybe_coroutine(check, ctx):
                return False
        for cooldown in self.cooldown:
            bucket = cooldown.bucket.get_bucket(ctx)
            retry_after = cooldown.get_retry_after(bucket, time())
            if retry_after:
                return False

        return True

    async def _call_before_invoke(self, ctx: Context) -> None:
        if self._before_hook is not None:
            await self._before_hook(ctx)
        if ctx.bot._before_hook is not None:
            await ctx.bot._before_hook(ctx)

    async def _call_after_invoke(self, ctx: Context) -> None:
        if self._after_hook is not None:
            await self._after_hook(ctx)
        if ctx.bot._after_hook is not None:
            await ctx.bot._after_hook(ctx)

    async def _parse_positional_or_keyword_argument(self, ctx: Context, param: inspect.Parameter, args: list) -> None:
        is_greedy = get_origin(param.annotation) is converters.Greedy
        greedy_args: list[Any] = []
        if ctx.lex.position == ctx.lex.end:
            args.append(await self._get_default(ctx, param))
        for argument in ctx.lex:
            try:
                transformed = await self._transform(ctx, param, argument)
            except BadArgument:
                if not is_greedy:
                    raise
                ctx.lex.undo()  # undo last read string for the next argument
                args.append(tuple(greedy_args))
                break
            if not is_greedy:
                args.append(transformed)
                break
            greedy_args.append(transformed)

    async def _parse_keyword_argument(self, ctx: Context, param: inspect.Parameter, kwargs: dict[str, Any]) -> None:
        kwargs[param.name] = await (  # kwarg only param denotes "consume rest" semantics
            self._transform(ctx, param, ctx.lex.rest) if ctx.lex.rest else self._get_default(ctx, param)
        )

    async def _parse_var_keyword_argument(self, ctx: Context, param: inspect.Parameter, kwargs: dict[str, Any]) -> None:
        kv_pairs = [arg.split("=") for arg in ctx.lex]
        if not kv_pairs:
            raise MissingRequiredArgument(param)  # defaults don't work here

        key_type, value_type = (
            (str, str) if param.annotation in (param.empty, dict) else get_args(param.annotation)
        )  # default to dict[str, str]

        key_converter = self._get_converter(key_type)
        value_converter = self._get_converter(value_type)
        try:
            for key_arg, value_arg in kv_pairs:
                if key_arg in kwargs:
                    raise DuplicateKeywordArgument(key_arg)
                kwargs.update(
                    {
                        await self._convert(ctx, key_converter, param, key_arg.strip()): await self._convert(
                            ctx, value_converter, param, value_arg.strip()
                        )
                    }
                )
        except ValueError:
            raise UnmatchedKeyValuePair("Unmatched key-value pair passed") from None

    async def _parse_var_position_argument(self, ctx: Context, param: inspect.Parameter, args: list) -> None:
        for arg in ctx.lex:
            transformed = await self._transform(ctx, param, arg)
            args.append(transformed)

    async def _parse_arguments(self, ctx: Context) -> None:
        args: list[Any] = []
        kwargs: dict[str, Any] = {}  # these are mutated by functions above

        for param in self.clean_params.values():
            if param.kind == param.POSITIONAL_OR_KEYWORD:
                await self._parse_positional_or_keyword_argument(ctx, param, args)
            elif param.kind == param.KEYWORD_ONLY:
                await self._parse_keyword_argument(ctx, param, kwargs)
                break
            elif param.kind == param.VAR_KEYWORD:  # same as **kwargs
                await self._parse_var_keyword_argument(ctx, param, kwargs)
                break
            elif param.kind == param.VAR_POSITIONAL:  # same as *args
                await self._parse_var_position_argument(ctx, param, args)
                break

        ctx.args = tuple(args)
        ctx.kwargs = kwargs

    def _transform(self, ctx: Context, param: inspect.Parameter, argument: str) -> Coroutine[None, None, Any]:
        parm_type = self._prepare_param(param)
        converter = self._get_converter(parm_type)
        return self._convert(ctx, converter, param, argument)

    def _prepare_param(self, param: inspect.Parameter) -> type:
        converter = param.annotation
        if converter is param.empty:
            if param.default is not param.empty:
                converter = str if param.default is None else type(param.default)
            else:
                converter = str
        return converter

    def _get_converter(self, param_type: type) -> converters.Converters:
        converters_ = converters.CONVERTERS.get(param_type, param_type)
        if isinstance(converters_, tuple):
            if len(converters_) == 1 or not self.special_converters:
                return converters_[0]
            for converter in converters_:
                try:
                    idx = self.special_converters.index(converter)
                except ValueError:
                    pass
                else:
                    return self.special_converters[idx]
            return converters_[0]
        return converters_

    async def _convert(
        self,
        ctx: Context,
        converter: converters.Converters,
        param: inspect.Parameter,
        argument: str,
    ) -> Any:
        if isinstance(converter, converters.ConverterBase):
            if isinstance(converter, type):  # needs to be instantiated
                converter = converter()
            try:
                return await converter.convert(ctx, argument)
            except Exception as exc:
                try:
                    name = converter.__name__
                except AttributeError:
                    name = converter.__class__.__name__
                raise BadArgument(f"{argument!r} failed to convert to {name}") from exc
        origin = get_origin(converter)
        if origin is not None:
            args = get_args(converter)
            for arg in args:
                converter = self._get_converter(arg)
                try:
                    ret = await self._convert(ctx, converter, param, argument)
                except BadArgument:
                    if origin is not Union:
                        raise
                else:
                    if origin is not Literal:
                        return ret
                    if arg == ret:
                        return ret

            if origin is Union and args[-1] is type(None):  # typing.Optional
                try:
                    return self._get_default(ctx, param)  # get the default if possible
                except MissingRequiredArgument:
                    return None  # fall back to None
            if origin is Literal:
                raise BadArgument(f"Expected one of {', '.join(args)} not {argument!r}")

            raise BadArgument(f"Failed to parse {argument!r} to any type")

        try:
            return converter(argument)
        except Exception as exc:
            try:
                name = converter.__name__
            except AttributeError:
                name = converter.__class__.__name__
            raise BadArgument(f"{argument!r} failed to convert to {name}") from exc

    async def _get_default(self, ctx: Context, param: inspect.Parameter) -> Any:
        if param.default is param.empty:
            raise MissingRequiredArgument(param)
        if isinstance(param.default, converters.Default):
            try:
                default = param.default() if isinstance(param.default, type) else param.default
                return await default.default(ctx)
            except Exception as exc:
                try:
                    name = param.default.__name__
                except AttributeError:
                    name = param.default.__class__.__name__
                raise BadArgument(f"{name} failed to return a default argument") from exc
        return param.default


class GroupMixin:
    """Mixin for something that can have commands registered under it.

    Attributes
    ----------
    case_insensitive
        Whether commands should be invoke-able case insensitively.
    """

    def __init__(self, *args: Any, **kwargs: Any):
        self.case_insensitive: bool = kwargs.get("case_insensitive", False)
        self.__commands__: dict[str, Command] = CaseInsensitiveDict() if self.case_insensitive else {}
        super().__init__(*args, **kwargs)

    @property
    def all_commands(self) -> list[Command]:
        """A list of the loaded commands."""
        return list(self.__commands__.values())

    @property
    def commands(self) -> set[Command]:
        """A set of the loaded commands without duplicates."""
        return set(self.__commands__.values())

    def add_command(self, command: Command) -> None:
        """Add a command to the internal commands list.

        Parameters
        ----------
        command
            The command to register.
        """
        if not isinstance(command, Command):
            raise TypeError("command should derive from commands.Command")

        if command.name in self.__commands__:
            raise ClientException(f"The command {command.name} is already registered.")

        for param in command.clean_params.values():
            if isinstance(param.annotation, str):
                raise ClientException(
                    f"Please rename the parameter {param.name} or make its annotation defined at runtime"
                )

        self.__commands__[command.name] = command
        for alias in command.aliases:
            if alias in self.__commands__:
                self.remove_command(command.name)
                raise ClientException(f"{alias} is already an existing command or alias.")
            self.__commands__[alias] = command

    def remove_command(self, name: str) -> Command | None:
        """Remove a command from the internal commands list.

        Parameters
        ----------
        name
            The name of the command to remove.

        Returns
        --------
        The removed command or ``None`` if the command was not found.
        """
        try:
            command = self.__commands__.pop(name)
        except ValueError:
            return None

        for alias in command.aliases:
            del self.__commands__[alias]
        return command

    def get_command(self, name: str) -> Command | None:
        """Get a command.

        Parameters
        ----------
        name
            The name of the command.

        Returns
        -------
        The found command or ``None``.
        """

        # fast path, no space in name
        if " " not in name:
            return self.__commands__.get(name)

        names = name.split()
        if not names:
            return None
        command = self.__commands__.get(names[0])
        if not isinstance(command, Group):
            return command

        for name in names[1:]:
            try:
                command = command.__commands__[name]
            except (AttributeError, KeyError):
                return None

        return command

    @overload
    def command(
        self,
        callback: CallbackType[P],
    ) -> Command[P]:
        ...

    @overload
    def command(self, callback: None) -> Callable[[CallbackType[P]], Command[P]]:
        ...

    @overload  # this also needs higher kinded types as cls should be type[C[P]] | None
    def command(
        self,
        *,
        name: str | None = ...,
        cls: type[C] | None = ...,
        help: str | None = ...,
        brief: str | None = ...,
        usage: str | None = ...,
        description: str | None = ...,
        aliases: Iterable[str] | None = ...,
        checks: list[CheckReturnType] = ...,
        cooldown: list[Cooldown] = ...,
        special_converters: list[type[converters.Converters]] = ...,
        cog: Cog | None = ...,
        parent: GroupMixin | None = ...,
        enabled: bool = ...,
        hidden: bool = ...,
        case_insensitive: bool = ...,
    ) -> Callable[[CallT], C]:
        ...

    def command(
        self,
        callback: CallT | None = None,
        *,
        name: str | None = None,
        cls: type[G] | None = None,
        **attrs: Any,
    ) -> Callable[[CallT], C] | C:
        """|maybecallabledeco|
        A decorator that invokes :func:`command` and adds the created :class:`Command` to the internal command list.

        Parameters
        ----------
        name
            The name of the command. Will default to ``callback.__name__``.
        cls: type[:class:`Command`]
            The class to construct the command from. Defaults to :class:`Command`.
        attrs
            The parameters to pass to the command's ``__init__``.

        Returns
        -------
        The created command.
        """

        def decorator(callback: CallT) -> C:
            attrs.setdefault("parent", self)
            result = command(callback, name=name, cls=cls or Command, **attrs)  # type: ignore
            self.add_command(result)
            return result

        return decorator(callback) if callback is not None else decorator

    @overload
    def group(
        self,
        callback: CallbackType[P],
    ) -> Group[P]:
        ...

    @overload
    def group(
        self,
        callback: None,
    ) -> Callable[[CallbackType[P]], Group[P]]:
        ...

    @overload
    def group(
        self,
        *,
        name: str | None = ...,
        cls: type[G] | None = ...,
        help: str | None = ...,
        brief: str | None = ...,
        usage: str | None = ...,
        description: str | None = ...,
        aliases: Iterable[str] | None = ...,
        checks: list[CheckReturnType] = ...,
        cooldown: list[Cooldown] = ...,
        special_converters: list[type[converters.Converters]] = ...,
        cog: Cog | None = ...,
        parent: GroupMixin | None = ...,
        enabled: bool = ...,
        hidden: bool = ...,
        case_insensitive: bool = ...,
    ) -> Callable[[CallT], G]:
        ...

    def group(
        self,
        callback: CallT | None = None,
        *,
        name: str | None = None,
        cls: type[G] | None = None,
        **attrs: Any,
    ) -> Callable[[CallT], G] | G:
        """|maybecallabledeco|
        A decorator that invokes :func:`group` and adds the created :class:`Group` to the internal command list.

        Parameters
        ----------
        name
            The name of the command. Will default to ``callback.__name__``.
        cls: type[:class:`Group`]
            The class to construct the command from. Defaults to :class:`Group`.
        attrs
            The parameters to pass to the command's ``__init__``.

        Returns
        -------
        The created group command.
        """

        def decorator(callback: CallbackType[P]) -> G:
            attrs.setdefault("parent", self)
            result = group(callback, name=name, cls=cls or Group, **attrs)  # type: ignore
            self.add_command(result)
            return result

        return decorator(callback) if callback is not None else decorator

    @property
    def children(self) -> list[Self]:
        """The commands children.

        Returns
        -------
        :class:`list`\\[:class:`Command`]
        """
        commands = []
        for command in self.commands:
            commands.append(command)
            if isinstance(command, Group):
                commands.extend(command.children)

        return commands

    def recursively_remove_all_commands(self) -> None:
        for command in self.commands:
            if isinstance(command, GroupMixin):
                command.recursively_remove_all_commands()
            self.remove_command(command.name)


class Group(GroupMixin, Command[P]):
    def __init__(self, func: CallbackType[P], **kwargs: Any):
        super().__init__(func, **kwargs)

    async def invoke(self, ctx: Context) -> None:
        command = self
        for command_name in ctx.lex:
            try:
                command = command.__commands__[command_name]
            except (AttributeError, KeyError):
                ctx.lex.undo()
                break

        await (super().invoke(ctx) if command is self else command.invoke(ctx))


@overload
def command(
    callback: CallbackType[P],
) -> Command[P]:
    ...


@overload
def command(callback: None) -> Callable[[CallbackType[P]], Command[P]]:
    ...


@overload
def command(
    *,
    name: str | None = ...,
    cls: type[C] | None = ...,
    help: str | None = ...,
    brief: str | None = ...,
    usage: str | None = ...,
    description: str | None = ...,
    aliases: Iterable[str] | None = ...,
    checks: list[CheckReturnType] = ...,
    cooldown: list[Cooldown] = ...,
    special_converters: list[type[converters.Converters]] = ...,
    cog: Cog | None = ...,
    parent: Command | None = ...,
    enabled: bool = ...,
    hidden: bool = ...,
    case_insensitive: bool = ...,
) -> Callable[[CallbackType], Command]:
    ...


def command(
    callback: CallbackType | None = None,
    *,
    name: str | None = None,
    cls: type[C] | None = None,
    **attrs: Any,
) -> Callable[[CallT], C] | C:
    """|maybecallabledeco|
    A decorator that turns a :term:`coroutine function` into a :class:`Command`.

    Parameters
    ----------
    name
        The name of the command. Will default to ``callback.__name__``.
    cls: type[:class:`Command`]
        The class to construct the command from. Defaults to :class:`Command`.
    attrs
        The attributes to pass to the command's ``__init__``.

    Returns
    -------
    The created command.
    """
    cls = cls or Command

    def decorator(callback: CallT) -> C:
        if isinstance(callback, Command):
            raise TypeError("callback is already a command.")
        return cls(callback, name=name, **attrs)

    return decorator(callback) if callback is not None else decorator


@overload
def group(
    callback: CallbackType[P],
) -> Group[P]:
    ...


@overload
def group(callback: None) -> Callable[[CallbackType[P]], Group[P]]:
    ...


@overload
def group(
    *,
    name: str | None = ...,
    cls: type[G] | None = ...,
    help: str | None = ...,
    brief: str | None = ...,
    usage: str | None = ...,
    description: str | None = ...,
    aliases: Iterable[str] | None = ...,
    checks: list[CheckReturnType] = ...,
    cooldown: list[Cooldown] = ...,
    special_converters: list[type[converters.Converters]] = ...,
    cog: Cog | None = ...,
    parent: Command | None = ...,
    enabled: bool = ...,
    hidden: bool = ...,
    case_insensitive: bool = ...,
) -> Callable[[CallT], G]:
    ...


def group(
    callback: CallT | None = None,
    *,
    name: str | None = None,
    cls: type[G] | None = None,
    **attrs: Any,
) -> Callable[[CallT], G] | G:
    """|maybecallabledeco|
    A decorator that turns a :term:`coroutine function` into a :class:`Group`.

    Parameters
    ----------
    name
        The name of the command. Will default to ``callback.__name__``.
    cls: type[:class:`Group`]
        The class to construct the command from. Defaults to :class:`Group`.
    attrs
        The attributes to pass to the command's ``__init__``.

    Returns
    -------
    The created group command.
    """

    return command(callback, name=name, cls=cls or Group, **attrs)


def check(predicate: CheckType) -> CheckReturnType:
    """
    A decorator that registers a function that *could be a* |coroutine_link|_ as a check to a command.

    They should take a singular argument representing the :class:`~steam.ext.commands.Context` for the message.

    Examples
    --------
    .. code-block:: python3

        def is_mod(ctx: commands.Context) -> bool:
            return ctx.clan and ctx.author in ctx.clan.mods


        @commands.check(is_mod)
        @bot.command
        async def kick(ctx: commands.Context, user: steam.User) -> None:
            ...

    This will raise an :exc:`steam.ext.commands.CheckFailure` if the user is not an a mod in the clan.

    Attributes
    ----------
    predicate
        The registered check, this will always be a :term:`coroutine function` even if the original check wasn't.
    """

    def decorator(func: MC) -> MC:
        if isinstance(func, Command):
            func.checks.append(predicate)
        else:
            if not hasattr(func, "__commands_checks__"):
                func.__commands_checks__ = []

            func.__commands_checks__.append(predicate)

        return func

    if inspect.iscoroutinefunction(predicate):
        decorator.predicate = predicate
    else:

        @functools.wraps(predicate)
        async def wrapper(ctx: Context) -> bool:
            return predicate(ctx)

        decorator.predicate = wrapper

    return decorator


@overload
def is_owner(command: None = ...) -> MCD:
    ...


@overload
def is_owner(command: MCD) -> MCD:
    ...


def is_owner(command: MCD | None = None) -> MCD:
    """|maybecallabledeco|
    A decorator that will only allow the bot's owner(s) to invoke the command.

    Warning
    -------
    This relies on :attr:`~steam.ext.commands.Bot.owner_id` or :attr:`~steam.ext.commands.Bot.owner_ids` to be set to
    function if they are not no one will be able to use these commands.
    """

    def predicate(ctx: Context) -> bool:
        if ctx.bot.owner_id and ctx.author.id64 == ctx.bot.owner_id:
            return True
        if ctx.bot.owner_ids and ctx.author.id64 in ctx.bot.owner_ids:
            return True
        raise NotOwner()

    decorator = check(predicate)
    return decorator(command) if command is not None else decorator


@overload
def dm_only(command: None = ...) -> MCD:
    ...


@overload
def dm_only(command: MCD) -> MCD:
    ...


def dm_only(command: MCD | None = None) -> MCD:
    """|maybecallabledeco|
    A decorator that will make a command only invokable in a :class:`steam.DMChannel`.
    """

    def predicate(ctx: Context) -> bool:
        if isinstance(ctx.channel, DMChannel):
            return True

        raise DMChannelOnly()

    decorator = check(predicate)
    return decorator(command) if command is not None else decorator


def cooldown(rate: int, per: float, type: BucketType = BucketType.Default) -> CommandDeco:
    """Give a :class:`Command`'s a cooldown.

    Parameters
    ----------
    rate
        The amount of times a command can be executed
        before being put on cooldown.
    per
        The amount of time to wait between cooldowns.
    type
        The bucket that the cooldown applies to.

    Examples
    --------
    Usage

    .. code-block:: python3

        @bot.command
        @commands.cooldown(rate=1, per=10, commands.BucketType.User)
        async def once_every_ten_seconds(ctx: commands.Context) -> None:
            ...  # this can only be invoked a user every ten seconds.
    """

    def decorator(command: MC) -> MC:
        if isinstance(command, Command):
            command.cooldown.append(Cooldown(rate, per, type))
        else:
            if not hasattr(command, "__commands_cooldown__"):
                command.__commands_cooldown__ = []
            command.__commands_cooldown__.append(Cooldown(rate, per, type))
        return command

    return decorator



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: steammessages_clientserver_userstats.proto
# plugin: python-betterproto
# Last updated 09/09/2021

from dataclasses import dataclass
from typing import List

import betterproto


@dataclass(eq=False, repr=False)
class CMsgClientGetUserStats(betterproto.Message):
    game_id: int = betterproto.fixed64_field(1)
    crc_stats: int = betterproto.uint32_field(2)
    schema_local_version: int = betterproto.int32_field(3)
    steam_id_for_user: int = betterproto.fixed64_field(4)


@dataclass(eq=False, repr=False)
class CMsgClientGetUserStatsResponse(betterproto.Message):
    game_id: int = betterproto.fixed64_field(1)
    eresult: int = betterproto.int32_field(2)
    crc_stats: int = betterproto.uint32_field(3)
    schema: bytes = betterproto.bytes_field(4)
    stats: List["CMsgClientGetUserStatsResponseStats"] = betterproto.message_field(5)
    achievement_blocks: List["CMsgClientGetUserStatsResponseAchievementBlocks"] = betterproto.message_field(6)


@dataclass(eq=False, repr=False)
class CMsgClientGetUserStatsResponseStats(betterproto.Message):
    stat_id: int = betterproto.uint32_field(1)
    stat_value: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientGetUserStatsResponseAchievementBlocks(betterproto.Message):
    achievement_id: int = betterproto.uint32_field(1)
    unlock_time: List[int] = betterproto.fixed32_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientStoreUserStatsResponse(betterproto.Message):
    game_id: int = betterproto.fixed64_field(1)
    eresult: int = betterproto.int32_field(2)
    crc_stats: int = betterproto.uint32_field(3)
    stats_failed_validation: List["CMsgClientStoreUserStatsResponseStatsFailedValidation"] = betterproto.message_field(
        4
    )
    stats_out_of_date: bool = betterproto.bool_field(5)


@dataclass(eq=False, repr=False)
class CMsgClientStoreUserStatsResponseStatsFailedValidation(betterproto.Message):
    stat_id: int = betterproto.uint32_field(1)
    reverted_stat_value: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientStoreUserStats2(betterproto.Message):
    game_id: int = betterproto.fixed64_field(1)
    settor_steam_id: int = betterproto.fixed64_field(2)
    settee_steam_id: int = betterproto.fixed64_field(3)
    crc_stats: int = betterproto.uint32_field(4)
    explicit_reset: bool = betterproto.bool_field(5)
    stats: List["CMsgClientStoreUserStats2Stats"] = betterproto.message_field(6)


@dataclass(eq=False, repr=False)
class CMsgClientStoreUserStats2Stats(betterproto.Message):
    stat_id: int = betterproto.uint32_field(1)
    stat_value: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientStatsUpdated(betterproto.Message):
    steam_id: int = betterproto.fixed64_field(1)
    game_id: int = betterproto.fixed64_field(2)
    crc_stats: int = betterproto.uint32_field(3)
    updated_stats: List["CMsgClientStatsUpdatedUpdatedStats"] = betterproto.message_field(4)


@dataclass(eq=False, repr=False)
class CMsgClientStatsUpdatedUpdatedStats(betterproto.Message):
    stat_id: int = betterproto.uint32_field(1)
    stat_value: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientStoreUserStats(betterproto.Message):
    game_id: int = betterproto.fixed64_field(1)
    explicit_reset: bool = betterproto.bool_field(2)
    stats_to_store: List["CMsgClientStoreUserStatsStatsToStore"] = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientStoreUserStatsStatsToStore(betterproto.Message):
    stat_id: int = betterproto.uint32_field(1)
    stat_value: int = betterproto.uint32_field(2)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: steammessages_friendmessages.proto
# plugin: python-betterproto
# Last updated 09/09/2021

from dataclasses import dataclass
from typing import List

import betterproto


class EMessageReactionType(betterproto.Enum):
    Invalid = 0
    Emoticon = 1
    Sticker = 2


@dataclass(eq=False, repr=False)
class GetRecentMessagesRequest(betterproto.Message):
    steamid1: int = betterproto.fixed64_field(1)
    steamid2: int = betterproto.fixed64_field(2)
    count: int = betterproto.uint32_field(3)
    most_recent_conversation: bool = betterproto.bool_field(4)
    rtime32_start_time: int = betterproto.fixed32_field(5)
    bbcode_format: bool = betterproto.bool_field(6)
    start_ordinal: int = betterproto.uint32_field(7)
    time_last: int = betterproto.uint32_field(8)
    ordinal_last: int = betterproto.uint32_field(9)


@dataclass(eq=False, repr=False)
class GetRecentMessagesResponse(betterproto.Message):
    messages: List["GetRecentMessagesResponseFriendMessage"] = betterproto.message_field(1)
    more_available: bool = betterproto.bool_field(4)


@dataclass(eq=False, repr=False)
class GetRecentMessagesResponseFriendMessage(betterproto.Message):
    accountid: int = betterproto.uint32_field(1)
    timestamp: int = betterproto.uint32_field(2)
    message: str = betterproto.string_field(3)
    ordinal: int = betterproto.uint32_field(4)
    reactions: List["GetRecentMessagesResponseFriendMessageMessageReaction"] = betterproto.message_field(5)


@dataclass(eq=False, repr=False)
class GetRecentMessagesResponseFriendMessageMessageReaction(betterproto.Message):
    reaction_type: "EMessageReactionType" = betterproto.enum_field(1)
    reaction: str = betterproto.string_field(2)
    reactors: List[int] = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class GetActiveMessageSessionsRequest(betterproto.Message):
    lastmessage_since: int = betterproto.uint32_field(1)
    only_sessions_with_messages: bool = betterproto.bool_field(2)


@dataclass(eq=False, repr=False)
class GetActiveMessageSessionsResponse(betterproto.Message):
    message_sessions: List["GetActiveMessageSessionsResponseFriendMessageSession"] = betterproto.message_field(1)
    timestamp: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class GetActiveMessageSessionsResponseFriendMessageSession(betterproto.Message):
    accountid_friend: int = betterproto.uint32_field(1)
    last_message: int = betterproto.uint32_field(2)
    last_view: int = betterproto.uint32_field(3)
    unread_message_count: int = betterproto.uint32_field(4)


@dataclass(eq=False, repr=False)
class SendMessageRequest(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)
    chat_entry_type: int = betterproto.int32_field(2)
    message: str = betterproto.string_field(3)
    contains_bbcode: bool = betterproto.bool_field(4)
    echo_to_sender: bool = betterproto.bool_field(5)
    low_priority: bool = betterproto.bool_field(6)
    client_message_id: str = betterproto.string_field(8)


@dataclass(eq=False, repr=False)
class SendMessageResponse(betterproto.Message):
    modified_message: str = betterproto.string_field(1)
    server_timestamp: int = betterproto.uint32_field(2)
    ordinal: int = betterproto.uint32_field(3)
    message_without_bb_code: str = betterproto.string_field(4)


@dataclass(eq=False, repr=False)
class AckMessageNotification(betterproto.Message):
    steamid_partner: int = betterproto.fixed64_field(1)
    timestamp: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class IsInFriendsUiBetaRequest(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)


@dataclass(eq=False, repr=False)
class IsInFriendsUiBetaResponse(betterproto.Message):
    online_in_friendsui: bool = betterproto.bool_field(1)
    has_used_friendsui: bool = betterproto.bool_field(2)


@dataclass(eq=False, repr=False)
class UpdateMessageReactionRequest(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)
    server_timestamp: int = betterproto.uint32_field(2)
    ordinal: int = betterproto.uint32_field(3)
    reaction_type: "EMessageReactionType" = betterproto.enum_field(4)
    reaction: str = betterproto.string_field(5)
    is_add: bool = betterproto.bool_field(6)


@dataclass(eq=False, repr=False)
class UpdateMessageReactionResponse(betterproto.Message):
    reactors: List[int] = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class IncomingMessageNotification(betterproto.Message):
    steamid_friend: int = betterproto.fixed64_field(1)
    chat_entry_type: int = betterproto.int32_field(2)
    from_limited_account: bool = betterproto.bool_field(3)
    message: str = betterproto.string_field(4)
    rtime32_server_timestamp: int = betterproto.fixed32_field(5)
    ordinal: int = betterproto.uint32_field(6)
    local_echo: bool = betterproto.bool_field(7)
    message_no_bbcode: str = betterproto.string_field(8)
    low_priority: bool = betterproto.bool_field(9)


@dataclass(eq=False, repr=False)
class MessageReactionNotification(betterproto.Message):
    steamid_friend: int = betterproto.fixed64_field(1)
    server_timestamp: int = betterproto.uint32_field(2)
    ordinal: int = betterproto.uint32_field(3)
    reactor: int = betterproto.fixed64_field(4)
    reaction_type: "EMessageReactionType" = betterproto.enum_field(5)
    reaction: str = betterproto.string_field(6)
    is_add: bool = betterproto.bool_field(7)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: content_manifest.proto
# plugin: python-betterproto
# Last updated 09/09/2021

from dataclasses import dataclass
from typing import List

import betterproto


class EContentDeltaChunkDataLocation(betterproto.Enum):
    InProtobuf = 0
    AfterProtobuf = 1


@dataclass(eq=False, repr=False)
class Payload(betterproto.Message):
    mappings: List["PayloadFileMapping"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class PayloadFileMapping(betterproto.Message):
    filename: str = betterproto.string_field(1)
    size: int = betterproto.uint64_field(2)
    flags: int = betterproto.uint32_field(3)
    sha_filename: bytes = betterproto.bytes_field(4)
    sha_content: bytes = betterproto.bytes_field(5)
    chunks: List["PayloadFileMappingChunkData"] = betterproto.message_field(6)
    linktarget: str = betterproto.string_field(7)


@dataclass(eq=False, repr=False)
class PayloadFileMappingChunkData(betterproto.Message):
    sha: bytes = betterproto.bytes_field(1)
    crc: int = betterproto.fixed32_field(2)
    offset: int = betterproto.uint64_field(3)
    cb_original: int = betterproto.uint32_field(4)
    cb_compressed: int = betterproto.uint32_field(5)


@dataclass(eq=False, repr=False)
class Metadata(betterproto.Message):
    depot_id: int = betterproto.uint32_field(1)
    gid_manifest: int = betterproto.uint64_field(2)
    creation_time: int = betterproto.uint32_field(3)
    filenames_encrypted: bool = betterproto.bool_field(4)
    cb_disk_original: int = betterproto.uint64_field(5)
    cb_disk_compressed: int = betterproto.uint64_field(6)
    unique_chunks: int = betterproto.uint32_field(7)
    crc_encrypted: int = betterproto.uint32_field(8)
    crc_clear: int = betterproto.uint32_field(9)


@dataclass(eq=False, repr=False)
class Signature(betterproto.Message):
    signature: bytes = betterproto.bytes_field(1)


@dataclass(eq=False, repr=False)
class ContentDeltaChunks(betterproto.Message):
    depot_id: int = betterproto.uint32_field(1)
    manifest_id_source: int = betterproto.uint64_field(2)
    manifest_id_target: int = betterproto.uint64_field(3)
    delta_chunks: List["ContentDeltaChunksDeltaChunk"] = betterproto.message_field(4)
    chunk_data_location: "EContentDeltaChunkDataLocation" = betterproto.enum_field(5)


@dataclass(eq=False, repr=False)
class ContentDeltaChunksDeltaChunk(betterproto.Message):
    sha_source: bytes = betterproto.bytes_field(1)
    sha_target: bytes = betterproto.bytes_field(2)
    size_original: int = betterproto.uint32_field(3)
    patch_method: int = betterproto.uint32_field(4)
    chunk: bytes = betterproto.bytes_field(5)
    size_delta: int = betterproto.uint32_field(6)



"""
The MIT License (MIT)

Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import TYPE_CHECKING, Generic, TypeVar

if TYPE_CHECKING:
    from .abc import Commentable
    from .message import Authors
    from .state import ConnectionState


__all__ = ("Comment",)

C = TypeVar("C", bound="Commentable")


@dataclass(repr=False)
class Comment(Generic[C]):
    """Represents a comment on a Steam profile.

    Attributes
    ----------
    id
        The comment's id.
    content
        The comment's content.
    author
        The author of the comment.
    created_at
        The time the comment was posted at.
    owner
        The comment sections owner.
    """

    __slots__ = ("content", "id", "created_at", "author", "owner", "_state")

    _state: ConnectionState
    id: int
    content: str
    created_at: datetime
    author: Authors
    owner: C

    def __repr__(self) -> str:
        attrs = ("id", "author")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<Comment {' '.join(resolved)}>"

    async def report(self) -> None:
        """Reports the comment."""
        await self._state.report_comment(self.owner, self.id)

    async def delete(self) -> None:
        """Deletes the comment."""
        await self._state.delete_comment(self.owner, self.id)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: steammessages_clientserver_lbs.proto
# plugin: python-betterproto
# Last updated 09/09/2021

from dataclasses import dataclass
from typing import List

import betterproto


@dataclass(eq=False, repr=False)
class CMsgClientLbsSetScore(betterproto.Message):
    app_id: int = betterproto.uint32_field(1)
    leaderboard_id: int = betterproto.int32_field(2)
    score: int = betterproto.int32_field(3)
    details: bytes = betterproto.bytes_field(4)
    upload_score_method: int = betterproto.int32_field(5)


@dataclass(eq=False, repr=False)
class CMsgClientLbsSetScoreResponse(betterproto.Message):
    eresult: int = betterproto.int32_field(1)
    leaderboard_entry_count: int = betterproto.int32_field(2)
    score_changed: bool = betterproto.bool_field(3)
    global_rank_previous: int = betterproto.int32_field(4)
    global_rank_new: int = betterproto.int32_field(5)


@dataclass(eq=False, repr=False)
class CMsgClientLbsSetUgc(betterproto.Message):
    app_id: int = betterproto.uint32_field(1)
    leaderboard_id: int = betterproto.int32_field(2)
    ugc_id: int = betterproto.fixed64_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientLbsSetUgcResponse(betterproto.Message):
    eresult: int = betterproto.int32_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientLbsFindOrCreateLb(betterproto.Message):
    app_id: int = betterproto.uint32_field(1)
    leaderboard_sort_method: int = betterproto.int32_field(2)
    leaderboard_display_type: int = betterproto.int32_field(3)
    create_if_not_found: bool = betterproto.bool_field(4)
    leaderboard_name: str = betterproto.string_field(5)


@dataclass(eq=False, repr=False)
class CMsgClientLbsFindOrCreateLbResponse(betterproto.Message):
    eresult: int = betterproto.int32_field(1)
    leaderboard_id: int = betterproto.int32_field(2)
    leaderboard_entry_count: int = betterproto.int32_field(3)
    leaderboard_sort_method: int = betterproto.int32_field(4)
    leaderboard_display_type: int = betterproto.int32_field(5)
    leaderboard_name: str = betterproto.string_field(6)


@dataclass(eq=False, repr=False)
class CMsgClientLbsGetLbEntries(betterproto.Message):
    app_id: int = betterproto.int32_field(1)
    leaderboard_id: int = betterproto.int32_field(2)
    range_start: int = betterproto.int32_field(3)
    range_end: int = betterproto.int32_field(4)
    leaderboard_data_request: int = betterproto.int32_field(5)
    steamids: List[int] = betterproto.fixed64_field(6)


@dataclass(eq=False, repr=False)
class CMsgClientLbsGetLbEntriesResponse(betterproto.Message):
    eresult: int = betterproto.int32_field(1)
    leaderboard_entry_count: int = betterproto.int32_field(2)
    entries: List["CMsgClientLbsGetLbEntriesResponseEntry"] = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientLbsGetLbEntriesResponseEntry(betterproto.Message):
    steam_id_user: int = betterproto.fixed64_field(1)
    global_rank: int = betterproto.int32_field(2)
    score: int = betterproto.int32_field(3)
    details: bytes = betterproto.bytes_field(4)
    ugc_id: int = betterproto.fixed64_field(5)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: steammessages_parental.steamclient.proto
# plugin: python-betterproto
# Last updated 09/09/2021

from dataclasses import dataclass
from typing import List

import betterproto


@dataclass(eq=False, repr=False)
class ParentalApp(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    is_allowed: bool = betterproto.bool_field(2)


@dataclass(eq=False, repr=False)
class ParentalSettings(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)
    applist_base_id: int = betterproto.uint32_field(2)
    applist_base_description: str = betterproto.string_field(3)
    applist_base: List["ParentalApp"] = betterproto.message_field(4)
    applist_custom: List["ParentalApp"] = betterproto.message_field(5)
    passwordhashtype: int = betterproto.uint32_field(6)
    salt: bytes = betterproto.bytes_field(7)
    passwordhash: bytes = betterproto.bytes_field(8)
    is_enabled: bool = betterproto.bool_field(9)
    enabled_features: int = betterproto.uint32_field(10)
    recovery_email: str = betterproto.string_field(11)
    is_site_license_lock: bool = betterproto.bool_field(12)


@dataclass(eq=False, repr=False)
class EnableParentalSettingsRequest(betterproto.Message):
    password: str = betterproto.string_field(1)
    settings: "ParentalSettings" = betterproto.message_field(2)
    sessionid: str = betterproto.string_field(3)
    enablecode: int = betterproto.uint32_field(4)
    steamid: int = betterproto.fixed64_field(10)


@dataclass(eq=False, repr=False)
class EnableParentalSettingsResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class DisableParentalSettingsRequest(betterproto.Message):
    password: str = betterproto.string_field(1)
    steamid: int = betterproto.fixed64_field(10)


@dataclass(eq=False, repr=False)
class DisableParentalSettingsResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class GetParentalSettingsRequest(betterproto.Message):
    steamid: int = betterproto.fixed64_field(10)


@dataclass(eq=False, repr=False)
class GetParentalSettingsResponse(betterproto.Message):
    settings: "ParentalSettings" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class GetSignedParentalSettingsRequest(betterproto.Message):
    priority: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class GetSignedParentalSettingsResponse(betterproto.Message):
    serialized_settings: bytes = betterproto.bytes_field(1)
    signature: bytes = betterproto.bytes_field(2)


@dataclass(eq=False, repr=False)
class SetParentalSettingsRequest(betterproto.Message):
    password: str = betterproto.string_field(1)
    settings: "ParentalSettings" = betterproto.message_field(2)
    new_password: str = betterproto.string_field(3)
    sessionid: str = betterproto.string_field(4)
    steamid: int = betterproto.fixed64_field(10)


@dataclass(eq=False, repr=False)
class SetParentalSettingsResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class ValidateTokenRequest(betterproto.Message):
    unlock_token: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class ValidateTokenResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class ValidatePasswordRequest(betterproto.Message):
    password: str = betterproto.string_field(1)
    session: str = betterproto.string_field(2)
    send_unlock_on_success: bool = betterproto.bool_field(3)


@dataclass(eq=False, repr=False)
class ValidatePasswordResponse(betterproto.Message):
    token: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class LockClientRequest(betterproto.Message):
    session: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class LockClientResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class RequestRecoveryCodeRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class RequestRecoveryCodeResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class DisableWithRecoveryCodeRequest(betterproto.Message):
    recovery_code: int = betterproto.uint32_field(1)
    steamid: int = betterproto.fixed64_field(10)


@dataclass(eq=False, repr=False)
class DisableWithRecoveryCodeResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class ParentalSettingsChangeNotification(betterproto.Message):
    serialized_settings: bytes = betterproto.bytes_field(1)
    signature: bytes = betterproto.bytes_field(2)
    password: str = betterproto.string_field(3)
    sessionid: str = betterproto.string_field(4)


@dataclass(eq=False, repr=False)
class ParentalUnlockNotification(betterproto.Message):
    password: str = betterproto.string_field(1)
    sessionid: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class ParentalLockNotification(betterproto.Message):
    sessionid: str = betterproto.string_field(1)
